from typing import Sequence

from eth_utils import to_tuple

from ssz.hashable_container import HashableContainer
from ssz.hashable_list import HashableList
from ssz.hashable_vector import HashableVector
from ssz.sedes import (
    BaseSedes,
    Bitlist,
    Bitvector,
    Boolean,
    ByteVector,
    Container,
    List,
    Serializable,
    UInt,
    Vector,
)

from .codec import DefaultCodec


def to_formatted_dict(value, sedes=None, codec=DefaultCodec):
    return dump(value, sedes, codec)


def dump(value, sedes=None, codec=DefaultCodec):
    if isinstance(value, Serializable):
        return dump_serializable(value, codec)
    elif isinstance(value, HashableContainer):
        return dump_hashable_container(value, codec)
    elif isinstance(value, (HashableList, HashableVector)):
        return dump_hashable_sequence(value, codec)
    elif sedes is None:
        raise ValueError("Sedes may only be omitted for serializables or hashables")

    if isinstance(sedes, Boolean):
        return dump_boolean(value, sedes, codec)
    elif isinstance(sedes, UInt):
        return dump_integer(value, sedes, codec)
    elif isinstance(sedes, ByteVector):
        return dump_bytes(value, sedes, codec)
    elif isinstance(sedes, List):
        return dump_list(value, sedes, codec)
    elif isinstance(sedes, Vector):
        return dump_vector(value, sedes, codec)
    elif isinstance(sedes, (Bitlist, Bitvector)):
        return dump_bits(value, sedes, codec)
    elif isinstance(sedes, Container):
        return dump_container(value, sedes, codec)
    elif isinstance(sedes, BaseSedes):
        raise Exception(
            f"Unreachable: All sedes types have been checked, {sedes} was not found"
        )
    else:
        raise TypeError(f"Expected BaseSedes, got {type(sedes)}")


def dump_boolean(value, sedes, codec):
    if not isinstance(value, bool):
        raise ValueError(f"Expected value of type bool, got {type(value)}")
    return codec.encode_bool(value, sedes)


def dump_integer(value, sedes, codec):
    return codec.encode_integer(value, sedes)


def dump_bytes(value, sedes, codec):
    return codec.encode_bytes(value, sedes)


def dump_list(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    return tuple(dump(element, sedes.element_sedes, codec) for element in value)


def dump_vector(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    if not len(value) == sedes.length:
        raise ValueError(f"Expected {sedes.length} elements, got {len(value)}")
    return tuple(dump(element, sedes.element_sedes, codec) for element in value)


def dump_bits(value, sedes, codec):
    return codec.encode_bytes(sedes.serialize(value), sedes)


@to_tuple
def dump_container(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    elif not len(value) == len(sedes.field_sedes):
        raise ValueError(
            f"Expected {len(sedes.field_sedes)} elements, got {len(value)}"
        )

    for element, element_sedes in zip(value, sedes.field_sedes):
        yield dump(element, element_sedes, codec)


def dump_serializable(value, codec):
    dumped_values = dump(tuple(value), value._meta.container_sedes)
    return dict(zip(value._meta.field_names, dumped_values))


def dump_hashable_container(value, codec):
    dumped_values = dump(tuple(value), value._meta.container_sedes)
    return {
        field_name: dumped_value
        for (field_name, _), dumped_value in zip(value._meta.fields, dumped_values)
    }


def dump_hashable_sequence(value, codec):
    return tuple(dump(element, value.sedes.element_sedes, codec) for element in value)



from typing import NamedTuple, Optional, Tuple

import ssz
from ssz.constants import SIGNATURE_FIELD_NAME
from ssz.sedes.base import BaseSedes
from ssz.sedes.container import Container
from ssz.sedes.serializable import BaseSerializable, MetaSerializable


class SignedMeta(NamedTuple):
    has_fields: bool
    fields: Optional[Tuple[Tuple[str, BaseSedes]]]
    container_sedes: Optional[Container]
    signed_container_sedes: Optional[Container]
    field_names: Optional[Tuple[str, ...]]
    field_attrs: Optional[Tuple[str, ...]]


class MetaSignedSerializable(MetaSerializable):
    def __new__(mcls, name, bases, namespace):
        cls = super().__new__(mcls, name, bases, namespace)

        if cls._meta.has_fields:
            if len(cls._meta.fields) < 2:
                raise TypeError(
                    f"Signed serializables need to have at least two fields"
                )
            if cls._meta.field_names[-1] != SIGNATURE_FIELD_NAME:
                raise TypeError(
                    f"Last field of signed serializable must be {SIGNATURE_FIELD_NAME}, but is "
                    f"{cls._meta.field_names[-1]}"
                )

            signed_container_sedes = Container(
                cls._meta.container_sedes.field_sedes[:-1]
            )
        else:
            signed_container_sedes = None

        meta = SignedMeta(
            has_fields=cls._meta.has_fields,
            fields=cls._meta.fields,
            container_sedes=cls._meta.container_sedes,
            signed_container_sedes=signed_container_sedes,
            field_names=cls._meta.field_names,
            field_attrs=cls._meta.field_attrs,
        )
        cls._meta = meta

        return cls


BaseSedes.register(MetaSignedSerializable)


class SignedSerializable(BaseSerializable, metaclass=MetaSignedSerializable):
    @property
    def signing_root(self):
        return ssz.get_hash_tree_root(self, self._meta.signed_container_sedes)



import functools
import itertools
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from eth_typing import Hash32
from eth_utils import to_dict, to_tuple
from eth_utils.toolz import groupby, partition, pipe
from pyrsistent import pvector
from pyrsistent._transformations import transform
from pyrsistent.typing import PVector

from ssz.abc import (
    HashableStructureAPI,
    HashableStructureEvolverAPI,
    ResizableHashableStructureAPI,
    ResizableHashableStructureEvolverAPI,
)
from ssz.constants import CHUNK_SIZE, ZERO_BYTES32
from ssz.hash_tree import HashTree
from ssz.sedes.base import BaseProperCompositeSedes

TStructure = TypeVar("TStructure", bound="BaseHashableStructure")
TResizableStructure = TypeVar(
    "TResizableStructure", bound="BaseResizableHashableStructure"
)
TElement = TypeVar("TElement")


def update_element_in_chunk(
    original_chunk: Hash32, index: int, element: bytes
) -> Hash32:
    """Replace part of a chunk with a given element.

    The chunk is interpreted as a concatenated sequence of equally sized elements. This function
    replaces the element given by its index in the chunk with the given data.

    If the length of the element is zero or not a divisor of the chunk size, a `ValueError` is
    raised. If the index is out of range, an `IndexError is raised.

    Example:
        >>> update_element_in_chunk(b"aabbcc", 1, b"xx")
        b'aaxxcc'
    """
    element_size = len(element)
    chunk_size = len(original_chunk)

    if element_size == 0:
        raise ValueError(f"Element size is zero")
    if chunk_size % element_size != 0:
        raise ValueError(f"Element size is not a divisor of chunk size: {element_size}")
    if not 0 <= index < chunk_size // element_size:
        raise IndexError(f"Index out of range for element size {element_size}: {index}")

    first_byte_index = index * element_size
    last_byte_index = first_byte_index + element_size

    prefix = original_chunk[:first_byte_index]
    suffix = original_chunk[last_byte_index:]
    return Hash32(prefix + element + suffix)


def update_elements_in_chunk(
    original_chunk: Hash32, updated_elements: Dict[int, bytes]
) -> Hash32:
    """Update multiple elements in a chunk.

    The set of updates is given by a dictionary mapping indices to elements. The items of the
    dictionary will be passed one by one to `update_element_in_chunk`.
    """
    return pipe(
        original_chunk,
        *(
            functools.partial(update_element_in_chunk, index=index, element=element)
            for index, element in updated_elements.items()
        ),
    )


def get_num_padding_elements(
    *, num_original_elements: int, num_original_chunks: int, element_size: int
) -> int:
    """Compute the number of elements that would still fit in the empty space of the last chunk."""
    total_size = num_original_chunks * CHUNK_SIZE
    used_size = num_original_elements * element_size
    padding_size = total_size - used_size
    num_elements_in_padding = padding_size // element_size
    return num_elements_in_padding


@to_dict
def get_updated_chunks(
    *,
    updated_elements: Dict[int, bytes],
    appended_elements: Sequence[bytes],
    original_chunks: Sequence[Hash32],
    element_size: int,
    num_original_elements: int,
    num_padding_elements: int,
) -> Generator[Tuple[int, Hash32], None, None]:
    """For an element changeset, compute the updates that have to be applied to the existing chunks.

    The changeset is given as a dictionary of element indices to updated elements and a sequence of
    appended elements. Note that appended elements that do not affect existing chunks are ignored.

    The pre-existing state is given by the sequence of original chunks and the number of elements
    represented by these chunks.

    The return value is a dictionary mapping chunk indices to chunks.
    """
    effective_appended_elements = appended_elements[:num_padding_elements]
    elements_per_chunk = CHUNK_SIZE // element_size

    padding_elements_with_indices = dict(
        enumerate(effective_appended_elements, start=num_original_elements)
    )
    effective_updated_elements = {**updated_elements, **padding_elements_with_indices}

    element_indices = effective_updated_elements.keys()
    element_indices_by_chunk = groupby(
        lambda element_index: element_index // elements_per_chunk, element_indices
    )

    for chunk_index, element_indices in element_indices_by_chunk.items():
        chunk_updates = {
            element_index
            % elements_per_chunk: effective_updated_elements[element_index]
            for element_index in element_indices
        }
        updated_chunk = update_elements_in_chunk(
            original_chunks[chunk_index], chunk_updates
        )
        yield chunk_index, updated_chunk


@to_tuple
def get_appended_chunks(
    *, appended_elements: Sequence[bytes], element_size: int, num_padding_elements: int
) -> Generator[Hash32, None, None]:
    """Get the sequence of appended chunks."""
    if len(appended_elements) <= num_padding_elements:
        return

    elements_per_chunk = CHUNK_SIZE // element_size

    chunk_partitioned_elements = partition(
        elements_per_chunk,
        appended_elements[num_padding_elements:],
        pad=b"\x00" * element_size,
    )
    for elements_in_chunk in chunk_partitioned_elements:
        yield Hash32(b"".join(elements_in_chunk))


class BaseHashableStructure(HashableStructureAPI[TElement]):
    def __init__(
        self,
        elements: PVector[TElement],
        hash_tree: HashTree,
        sedes: BaseProperCompositeSedes,
        max_length: Optional[int] = None,
    ) -> None:
        self._elements = elements
        self._hash_tree = hash_tree
        self._sedes = sedes
        self._max_length = max_length

    @classmethod
    def from_iterable_and_sedes(
        cls,
        iterable: Iterable[TElement],
        sedes: BaseProperCompositeSedes,
        max_length: Optional[int] = None,
    ):
        elements = pvector(iterable)
        if max_length and len(elements) > max_length:
            raise ValueError(
                f"Number of elements {len(elements)} exceeds maximum length {max_length}"
            )

        serialized_elements = [
            sedes.serialize_element_for_tree(index, element)
            for index, element in enumerate(elements)
        ]
        appended_chunks = get_appended_chunks(
            appended_elements=serialized_elements,
            element_size=sedes.element_size_in_tree,
            num_padding_elements=0,
        )
        hash_tree = HashTree.compute(
            appended_chunks or [ZERO_BYTES32], sedes.chunk_count
        )
        return cls(elements, hash_tree, sedes, max_length)

    @property
    def elements(self) -> PVector[TElement]:
        return self._elements

    @property
    def hash_tree(self) -> HashTree:
        return self._hash_tree

    @property
    def chunks(self) -> PVector[Hash32]:
        return self.hash_tree.chunks

    @property
    def max_length(self) -> Optional[int]:
        return self._max_length

    @property
    def raw_root(self) -> Hash32:
        return self.hash_tree.root

    @property
    def sedes(self) -> BaseProperCompositeSedes:
        return self._sedes

    #
    # Hash and equality
    #
    def __hash__(self) -> int:
        # hashable structures have the same hash if they share both sedes and root
        return hash((self.sedes, self.hash_tree_root))

    def __eq__(self, other: Any) -> bool:
        # hashable structures are equal if they use the same sedes and have the same root
        if isinstance(other, BaseHashableStructure):
            sedes_equal = self.sedes == other.sedes
            roots_equal = self.hash_tree_root == other.hash_tree_root
            return sedes_equal and roots_equal
        else:
            return False

    #
    # PVector interface
    #
    def __len__(self) -> int:
        return len(self.elements)

    def __getitem__(self, index: int) -> TElement:
        return self.elements[index]

    def __iter__(self) -> Iterator[TElement]:
        return iter(self.elements)

    def transform(self, *transformations):
        return transform(self, transformations)

    def mset(self: TStructure, *args: Union[int, TElement]) -> TStructure:
        if len(args) % 2 != 0:
            raise TypeError(
                f"mset must be called with an even number of arguments, got {len(args)}"
            )

        evolver = self.evolver()
        for index, value in partition(2, args):
            evolver[index] = value
        return evolver.persistent()

    def set(self: TStructure, index: int, value: TElement) -> TStructure:
        return self.mset(index, value)

    def evolver(
        self: TStructure
    ) -> "HashableStructureEvolverAPI[TStructure, TElement]":
        return HashableStructureEvolver(self)


class HashableStructureEvolver(HashableStructureEvolverAPI[TStructure, TElement]):
    def __init__(self, hashable_structure: TStructure) -> None:
        self._original_structure = hashable_structure
        self._updated_elements: Dict[int, TElement] = {}
        # `self._appended_elements` is only used in the subclass ResizableHashableStructureEvolver,
        # but the implementation of `persistent` already processes it so that it does not have to
        # be implemented twice.
        self._appended_elements: List[TElement] = []

    def __getitem__(self, index: int) -> TElement:
        if index < 0:
            index += len(self)

        if index in self._updated_elements:
            return self._updated_elements[index]
        elif 0 <= index < len(self._original_structure):
            return self._original_structure[index]
        elif 0 <= index < len(self):
            return self._appended_elements[index - len(self._original_structure)]
        else:
            raise IndexError(f"Index out of bounds: {index}")

    def set(self, index: int, element: TElement) -> None:
        self[index] = element

    def __setitem__(self, index: int, element: TElement) -> None:
        if index < 0:
            index += len(self)

        if 0 <= index < len(self._original_structure):
            self._updated_elements[index] = element
        elif 0 <= index < len(self):
            self._appended_elements[index - len(self._original_structure)] = element
        else:
            raise IndexError(f"Index out of bounds: {index}")

    def __len__(self) -> int:
        return len(self._original_structure) + len(self._appended_elements)

    def is_dirty(self) -> bool:
        return bool(self._updated_elements or self._appended_elements)

    def persistent(self) -> TStructure:
        if not self.is_dirty():
            return self._original_structure

        sedes = self._original_structure.sedes

        num_original_elements = len(self._original_structure)
        num_original_chunks = len(self._original_structure.chunks)
        num_padding_elements = get_num_padding_elements(
            num_original_elements=num_original_elements,
            num_original_chunks=num_original_chunks,
            element_size=sedes.element_size_in_tree,
        )

        updated_elements = {
            index: sedes.serialize_element_for_tree(index, element)
            for index, element in self._updated_elements.items()
        }
        appended_elements = [
            sedes.serialize_element_for_tree(index, element)
            for index, element in enumerate(
                self._appended_elements, start=num_original_elements
            )
        ]

        updated_chunks = get_updated_chunks(
            updated_elements=updated_elements,
            appended_elements=appended_elements,
            original_chunks=self._original_structure.chunks,
            num_original_elements=num_original_elements,
            num_padding_elements=num_padding_elements,
            element_size=sedes.element_size_in_tree,
        )
        appended_chunks = get_appended_chunks(
            appended_elements=appended_elements,
            element_size=sedes.element_size_in_tree,
            num_padding_elements=num_padding_elements,
        )

        elements = self._original_structure.elements.mset(
            *itertools.chain.from_iterable(  # type: ignore
                self._updated_elements.items()
            )
        ).extend(self._appended_elements)
        hash_tree = self._original_structure.hash_tree.mset(
            *itertools.chain.from_iterable(  # type: ignore
                updated_chunks.items()
            )
        ).extend(appended_chunks)

        return self._original_structure.__class__(
            elements, hash_tree, self._original_structure.sedes
        )


class BaseResizableHashableStructure(
    BaseHashableStructure, ResizableHashableStructureAPI[TElement]
):
    def append(self: TResizableStructure, value: TElement) -> TResizableStructure:
        evolver = self.evolver()
        evolver.append(value)
        return evolver.persistent()

    def extend(
        self: TResizableStructure, values: Iterable[TElement]
    ) -> TResizableStructure:
        evolver = self.evolver()
        evolver.extend(values)
        return evolver.persistent()

    def __add__(
        self: TResizableStructure, values: Iterable[TElement]
    ) -> TResizableStructure:
        return self.extend(values)

    def __mul__(self: TResizableStructure, times: int) -> TResizableStructure:
        if times <= 0:
            raise ValueError("Multiplication factor must be positive: {times}")
        elif times == 1:
            return self
        else:
            return (self + self) * (times - 1)

    def evolver(
        self: TResizableStructure,
    ) -> "ResizableHashableStructureEvolverAPI[TResizableStructure, TElement]":
        return ResizableHashableStructureEvolver(self)


class ResizableHashableStructureEvolver(
    HashableStructureEvolver, ResizableHashableStructureEvolverAPI[TStructure, TElement]
):
    def append(self, element: TElement) -> None:
        max_length = self._original_structure.max_length
        if max_length is not None and len(self) + 1 > max_length:
            raise ValueError("Structure would exceed maximum length {max_length}")
        self._appended_elements.append(element)

    def extend(self, elements: Iterable[TElement]) -> None:
        extension = list(elements)

        max_length = self._original_structure.max_length
        if max_length is not None and len(self) + len(extension) > max_length:
            raise ValueError("Structure would exceed maximum length {max_length}")

        self._appended_elements.extend(extension)



from typing import Any

from ssz.exceptions import DeserializationError, SerializationError
from ssz.sedes.basic import BasicSedes


class Byte(BasicSedes[bytes, bytes]):
    size = 1

    def __init__(self) -> None:
        super().__init__(1)

    def serialize(self, value: bytes) -> bytes:
        if len(value) != 1:
            raise SerializationError(
                f"The `Byte` sedes can only serialize single bytes.  Got: {value!r}"
            )
        return value

    def deserialize(self, data: bytes) -> bytes:
        if len(data) != 1:
            raise DeserializationError(
                f"The `Byte` sedes can only deserialize single bytes.  Got: {data!r}"
            )
        return data

    def get_sedes_id(self) -> str:
        return self.__class__.__name__

    def __hash__(self) -> int:
        return hash((Byte,))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Byte)


byte = Byte()



from typing import IO, Any, Iterable, Sequence, Tuple

from eth_typing import Hash32
from eth_utils import ValidationError, to_tuple
from eth_utils.toolz import sliding_window

from ssz.exceptions import DeserializationError, SerializationError
from ssz.hashable_structure import BaseHashableStructure
from ssz.sedes.base import BaseSedes, TSedes
from ssz.sedes.basic import ProperCompositeSedes
from ssz.typing import CacheObj
from ssz.utils import merkleize, read_exact, s_decode_offset


@to_tuple
def _deserialize_fixed_size_items_and_offsets(stream, field_sedes):
    for sedes in field_sedes:
        if sedes.is_fixed_sized:
            field_size = sedes.get_fixed_size()
            field_data = read_exact(field_size, stream)
            yield (sedes.deserialize(field_data), sedes)
        else:
            yield (s_decode_offset(stream), sedes)


class Container(ProperCompositeSedes[Sequence[Any], Tuple[Any, ...]]):
    def __init__(self, field_sedes: Sequence[TSedes]) -> None:
        if len(field_sedes) == 0:
            raise ValidationError("Cannot define container without any fields")
        self.field_sedes = tuple(field_sedes)

    #
    # Size
    #
    @property
    def is_fixed_sized(self):
        return all(field.is_fixed_sized for field in self.field_sedes)

    def get_fixed_size(self):
        if not self.is_fixed_sized:
            raise ValueError("Container contains dynamically sized elements")

        return sum(field.get_fixed_size() for field in self.field_sedes)

    #
    # Serialization
    #
    def get_element_sedes(self, index: int) -> BaseSedes:
        return self.field_sedes[index]

    @property
    def is_packing(self) -> bool:
        return False

    @property
    def chunk_count(self) -> int:
        return len(self.field_sedes)

    def _validate_serializable(self, value: Sequence[Any]) -> None:
        if len(value) != len(self.field_sedes):
            raise SerializationError(
                f"Incorrect element count: Expected: {len(self.field_sedes)} / Got: {len(value)}"
            )

    #
    # Deserialization
    #
    def deserialize_fixed_size_parts(
        self, stream: IO[bytes]
    ) -> Iterable[Tuple[Tuple[Any], Tuple[int, TSedes]]]:
        fixed_items_and_offets = _deserialize_fixed_size_items_and_offsets(
            stream, self.field_sedes
        )
        fixed_size_values = tuple(
            item for item, sedes in fixed_items_and_offets if sedes.is_fixed_sized
        )
        offset_pairs = tuple(
            (item, sedes)
            for item, sedes in fixed_items_and_offets
            if not sedes.is_fixed_sized
        )
        return fixed_size_values, offset_pairs

    @to_tuple
    def deserialize_variable_size_parts(
        self, offset_pairs: Tuple[Tuple[int, TSedes], ...], stream: IO[bytes]
    ) -> Iterable[Any]:
        offsets, fields = zip(*offset_pairs)

        *head_fields, last_field = fields
        for sedes, (left_offset, right_offset) in zip(
            head_fields, sliding_window(2, offsets)
        ):
            field_length = right_offset - left_offset
            field_data = read_exact(field_length, stream)
            yield sedes.deserialize(field_data)

        # simply reading to the end of the current stream gives us all of the final element data
        final_field_data = stream.read()
        yield last_field.deserialize(final_field_data)

    def _deserialize_stream(self, stream: IO[bytes]) -> Tuple[Any, ...]:
        if not self.field_sedes:
            # TODO: likely remove once
            # https://github.com/ethereum/eth2.0-specs/issues/854 is resolved
            return tuple()

        fixed_size_values, offset_pairs = self.deserialize_fixed_size_parts(stream)

        if not offset_pairs:
            return fixed_size_values

        variable_size_values = self.deserialize_variable_size_parts(
            offset_pairs, stream
        )

        fixed_size_parts_iter = iter(fixed_size_values)
        variable_size_parts_iter = iter(variable_size_values)

        value = tuple(
            next(fixed_size_parts_iter)
            if sedes.is_fixed_sized
            else next(variable_size_parts_iter)
            for sedes in self.field_sedes
        )

        # Verify that both iterables have been fully consumed.
        try:
            next(fixed_size_parts_iter)
        except StopIteration:
            pass
        else:
            raise DeserializationError("Did not consume all fixed size values")

        try:
            next(variable_size_parts_iter)
        except StopIteration:
            pass
        else:
            raise DeserializationError("Did not consume all variable size values")

        return value

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: Tuple[Any, ...]) -> bytes:
        if isinstance(value, BaseHashableStructure) and value.sedes == self:
            return value.hash_tree_root

        merkle_leaves = tuple(
            sedes.get_hash_tree_root(element)
            for element, sedes in zip(value, self.field_sedes)
        )
        return merkleize(merkle_leaves)

    def get_hash_tree_root_and_leaves(
        self, value: Tuple[Any, ...], cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        merkle_leaves = ()
        for element, sedes in zip(value, self.field_sedes):
            key = sedes.get_key(element)
            if key not in cache:
                if hasattr(sedes, "get_hash_tree_root_and_leaves"):
                    root, cache = sedes.get_hash_tree_root_and_leaves(element, cache)
                    cache[key] = root
                else:
                    cache[key] = sedes.get_hash_tree_root(element)

            merkle_leaves += (cache[key],)

        return merkleize(merkle_leaves), cache

    def serialize(self, value) -> bytes:
        if hasattr(value, "_serialize_cache") and value._serialize_cache is not None:
            return value._serialize_cache
        elif hasattr(value, "_serialize_cache") and value._serialize_cache is None:
            value._serialize_cache = super().serialize(value)
            return value._serialize_cache
        else:
            return super().serialize(value)

    def get_sedes_id(self) -> str:
        return ",".join(field.get_sedes_id() for field in self.field_sedes)

    #
    # Equality and hashing
    #
    def __hash__(self) -> int:
        return hash((hash(Container), hash(self.field_sedes)))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Container) and other.field_sedes == self.field_sedes



from typing import IO, Any, Iterable, Sequence, Tuple

from eth_typing import Hash32
from eth_utils import to_tuple
from eth_utils.toolz import cons, sliding_window

from ssz.cache.utils import (
    get_merkle_leaves_with_cache,
    get_merkle_leaves_without_cache,
)
from ssz.constants import OFFSET_SIZE
from ssz.exceptions import DeserializationError
from ssz.hashable_list import HashableList
from ssz.hashable_structure import BaseHashableStructure
from ssz.sedes.base import BaseSedes
from ssz.sedes.basic import BasicSedes, HomogeneousProperCompositeSedes
from ssz.typing import CacheObj, TDeserialized, TSerializable
from ssz.utils import (
    merkleize,
    merkleize_with_cache,
    mix_in_length,
    pack,
    read_exact,
    s_decode_offset,
)

TSedesPairs = Tuple[Tuple[BaseSedes[TSerializable, TDeserialized], TSerializable], ...]


class List(
    HomogeneousProperCompositeSedes[Sequence[TSerializable], Tuple[TDeserialized, ...]]
):
    #
    # Size
    #
    is_fixed_sized = False

    def get_fixed_size(self):
        raise ValueError("List has no static size")

    #
    # Deserialization
    #
    def get_element_sedes(self, index) -> BaseSedes[TSerializable, TDeserialized]:
        return self.element_sedes

    def _deserialize_stream(self, stream: IO[bytes]) -> HashableList[TDeserialized]:
        elements = self._deserialize_stream_to_tuple(stream)
        return HashableList.from_iterable(elements, sedes=self)

    @to_tuple
    def _deserialize_stream_to_tuple(
        self, stream: IO[bytes]
    ) -> Iterable[TDeserialized]:
        if self.element_sedes.is_fixed_sized:
            element_size = self.element_sedes.get_fixed_size()
            data = stream.read()
            if len(data) % element_size != 0:
                raise DeserializationError(
                    f"Invalid max_length. List is comprised of a fixed size sedes "
                    f"but total serialized data is not an even multiple of the "
                    f"element size. data max_length: {len(data)}  element size: "
                    f"{element_size}"
                )
            for start_idx in range(0, len(data), element_size):
                segment = data[start_idx : start_idx + element_size]
                yield self.element_sedes.deserialize(segment)
        else:
            stream_zero_loc = stream.tell()
            try:
                first_offset = s_decode_offset(stream)
            except DeserializationError:
                if stream.tell() == stream_zero_loc:
                    # Empty list
                    return
                else:
                    raise

            num_remaining_offset_bytes = first_offset - stream.tell()
            if num_remaining_offset_bytes % OFFSET_SIZE != 0:
                raise DeserializationError(
                    f"Offset bytes was not a multiple of {OFFSET_SIZE}.  Got "
                    f"{num_remaining_offset_bytes}"
                )

            num_remaining_offsets = num_remaining_offset_bytes // OFFSET_SIZE
            tail_offsets = tuple(
                s_decode_offset(stream) for _ in range(num_remaining_offsets)
            )

            offsets = tuple(cons(first_offset, tail_offsets))

            for left_offset, right_offset in sliding_window(2, offsets):
                element_length = right_offset - left_offset
                element_data = read_exact(element_length, stream)
                yield self.element_sedes.deserialize(element_data)

            # simply reading to the end of the current stream gives us all of the final element data
            final_element_data = stream.read()
            yield self.element_sedes.deserialize(final_element_data)

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: Iterable[TSerializable]) -> bytes:
        if isinstance(value, BaseHashableStructure) and value.sedes == self:
            return value.hash_tree_root

        if isinstance(self.element_sedes, BasicSedes):
            serialized_items = tuple(
                self.element_sedes.serialize(element) for element in value
            )
            merkle_leaves = pack(serialized_items)
        else:
            merkle_leaves = tuple(
                self.element_sedes.get_hash_tree_root(element) for element in value
            )

        return mix_in_length(
            merkleize(merkle_leaves, limit=self.chunk_count), len(value)
        )

    def get_hash_tree_root_and_leaves(
        self, value: TSerializable, cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        merkle_leaves = ()
        if isinstance(self.element_sedes, BasicSedes):
            serialized_items = tuple(
                self.element_sedes.serialize(element) for element in value
            )
            merkle_leaves = pack(serialized_items)
        else:
            has_get_hash_tree_root_and_leaves = hasattr(
                self.element_sedes, "get_hash_tree_root_and_leaves"
            )
            if has_get_hash_tree_root_and_leaves:
                merkle_leaves = get_merkle_leaves_with_cache(
                    value, self.element_sedes, cache
                )
            else:
                merkle_leaves = get_merkle_leaves_without_cache(
                    value, self.element_sedes
                )

        merkleize_result, cache = merkleize_with_cache(
            merkle_leaves, cache=cache, limit=self.chunk_count
        )
        return mix_in_length(merkleize_result, len(value)), cache

    #
    # Equality and hashing
    #
    def __hash__(self) -> int:
        return hash((hash(List), hash(self.element_sedes), self.max_length))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, List)
            and other.element_sedes == self.element_sedes
            and other.max_length == self.max_length
        )



import abc
import collections
import copy
import operator
import re
from typing import NamedTuple, Optional, Sequence, Tuple, Type

from eth_utils import ValidationError, to_dict, to_set, to_tuple
from eth_utils.toolz import assoc, merge
from lru import LRU

from ssz.cache.cache import DEFAULT_CACHE_SIZE
from ssz.cache.utils import get_base_key
from ssz.constants import FIELDS_META_ATTR
from ssz.sedes.base import BaseSedes
from ssz.sedes.container import Container
from ssz.typing import TSerializable
from ssz.utils import get_duplicates, is_immutable_field_value


class Meta(NamedTuple):
    has_fields: bool
    fields: Optional[Tuple[Tuple[str, BaseSedes], ...]]
    container_sedes: Optional[Container]
    field_names: Optional[Tuple[str, ...]]
    field_attrs: Optional[Tuple[str, ...]]


def validate_args_and_kwargs(args, kwargs, arg_names):
    duplicate_arg_names = get_duplicates(arg_names)
    if duplicate_arg_names:
        raise ValueError(
            "Duplicate argument names: {0}".format(sorted(duplicate_arg_names))
        )

    needed_arg_names = set(arg_names[len(args) :])
    used_arg_names = set(arg_names[: len(args)])

    duplicate_arg_names = used_arg_names.intersection(kwargs.keys())
    if duplicate_arg_names:
        raise TypeError("Duplicate kwargs: {0}".format(sorted(duplicate_arg_names)))

    unknown_arg_names = set(kwargs.keys()).difference(arg_names)
    if unknown_arg_names:
        raise TypeError("Unknown kwargs: {0}".format(sorted(unknown_arg_names)))

    missing_arg_names = set(needed_arg_names).difference(kwargs.keys())
    if missing_arg_names:
        raise TypeError("Missing kwargs: {0}".format(sorted(missing_arg_names)))


@to_tuple
def merge_kwargs_to_args(args, kwargs, arg_names):
    validate_args_and_kwargs(args, kwargs, arg_names)

    needed_arg_names = arg_names[len(args) :]

    yield from args
    for arg_name in needed_arg_names:
        yield kwargs[arg_name]


@to_dict
def merge_args_to_kwargs(args, kwargs, arg_names):
    yield from kwargs.items()
    for value, name in zip(args, arg_names):
        yield name, value


class BaseSerializable(collections.Sequence):
    cache = None

    def __init__(self, *args, cache=None, **kwargs):
        arg_names = self._meta.field_names or ()
        validate_args_and_kwargs(args, kwargs, arg_names)
        field_values = merge_kwargs_to_args(args, kwargs, arg_names)

        # Ensure that all the fields have been given values in initialization
        if len(field_values) != len(arg_names):
            raise TypeError(
                "Argument count mismatch. expected {0} - got {1} - missing {2}".format(
                    len(arg_names),
                    len(field_values),
                    ",".join(arg_names[len(field_values) :]),
                )
            )

        for value, attr in zip(field_values, self._meta.field_attrs or ()):
            setattr(self, attr, make_immutable(value))

        self.cache = LRU(DEFAULT_CACHE_SIZE) if cache is None else cache

    def as_dict(self):
        return dict(
            (field, value) for field, value in zip(self._meta.field_names, self)
        )

    def __iter__(self):
        for attr in self._meta.field_attrs:
            yield getattr(self, attr)

    def __getitem__(self, index):
        if isinstance(index, int):
            attr = self._meta.field_attrs[index]
            return getattr(self, attr)
        elif isinstance(index, slice):
            field_slice = self._meta.field_attrs[index]
            return tuple(getattr(self, field) for field in field_slice)
        elif isinstance(index, str):
            return getattr(self, index)
        else:
            raise IndexError(
                "Unsupported type for __getitem__: {0}".format(type(index))
            )

    def __len__(self):
        return len(self._meta.fields)

    def __eq__(self, other):
        satisfies_class_relationship = issubclass(
            self.__class__, other.__class__
        ) or issubclass(other.__class__, self.__class__)

        if not satisfies_class_relationship:
            return False
        else:
            return self.hash_tree_root == other.hash_tree_root

    def __getstate__(self):
        state = self.__dict__.copy()
        # The hash() builtin is not stable across processes
        # (https://docs.python.org/3/reference/datamodel.html#object.__hash__), so we do this here
        # to ensure pickled instances don't carry the cached hash() as that may cause issues like
        # https://github.com/ethereum/py-evm/issues/1318
        state["_hash_cache"] = None
        return state

    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is None:
            self._hash_cache = hash(self.__class__) * int.from_bytes(
                self.hash_tree_root, "little"
            )
        return self._hash_cache

    def copy(self, *args, **kwargs):
        missing_overrides = (
            set(self._meta.field_names)
            .difference(kwargs.keys())
            .difference(self._meta.field_names[: len(args)])
        )
        unchanged_kwargs = {
            key: value if is_immutable_field_value(value) else copy.deepcopy(value)
            for key, value in self.as_dict().items()
            if key in missing_overrides
        }

        combined_kwargs = dict(**unchanged_kwargs, **kwargs)
        all_kwargs = merge_args_to_kwargs(args, combined_kwargs, self._meta.field_names)

        result = type(self)(**all_kwargs)
        result.cache = self.cache

        return result

    def reset_cache(self):
        self.cache.clear()
        self._fixed_size_section_length_cache = None
        self._serialize_cache = None

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}

        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result

        for k, v in self.__dict__.items():
            if k != "cache":
                setattr(result, k, copy.deepcopy(v, memodict))

        result.cache = self.cache
        result._fixed_size_section_length_cache = self._fixed_size_section_length_cache

        return result

    _fixed_size_section_length_cache = None
    _serialize_cache = None

    @property
    def hash_tree_root(self):
        return self.__class__.get_hash_tree_root(self, cache=True)

    @classmethod
    def get_sedes_id(cls) -> str:
        # Serializable implementation name should be unique
        return cls.__name__

    def get_key(self) -> bytes:
        # Serilaize with self._meta.container_sedes
        key = get_base_key(self._meta.container_sedes, self).hex()

        if len(key) == 0:
            key = ""

        return f"{self.__class__.get_sedes_id()}{key}"


def make_immutable(value):
    if isinstance(value, list):
        return tuple(make_immutable(item) for item in value)
    else:
        return value


@to_tuple
def _mk_field_attrs(field_names, extra_namespace):
    namespace = set(field_names).union(extra_namespace)
    for field in field_names:
        while True:
            field = "_" + field
            if field not in namespace:
                namespace.add(field)
                yield field
                break


@to_dict
def _mk_field_props(field_names, field_attrs):
    for field, attr in zip(field_names, field_attrs):
        getter = operator.attrgetter(attr)
        yield field, property(getter)


def _validate_field_names(field_names: Sequence[str]) -> None:
    # check that field names are unique
    duplicate_field_names = get_duplicates(field_names)
    if duplicate_field_names:
        raise TypeError(
            "The following fields are duplicated in the `fields` "
            "declaration: "
            "{0}".format(",".join(sorted(duplicate_field_names)))
        )

    # check that field names are valid identifiers
    invalid_field_names = {
        field_name for field_name in field_names if not _is_valid_identifier(field_name)
    }
    if invalid_field_names:
        raise TypeError(
            "The following field names are not valid python identifiers: {0}".format(
                ",".join("`{0}`".format(item) for item in sorted(invalid_field_names))
            )
        )


IDENTIFIER_REGEX = re.compile(r"^[^\d\W]\w*\Z", re.UNICODE)


def _is_valid_identifier(value):
    # Source: https://stackoverflow.com/questions/5474008/regular-expression-to-confirm-whether-a-string-is-a-valid-identifier-in-python  # noqa: E501
    if not isinstance(value, str):
        return False
    return bool(IDENTIFIER_REGEX.match(value))


@to_set
def _get_class_namespace(cls):
    if hasattr(cls, "__dict__"):
        yield from cls.__dict__.keys()
    if hasattr(cls, "__slots__"):
        yield from cls.__slots__


class MetaSerializable(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace):
        declares_fields = FIELDS_META_ATTR in namespace

        if declares_fields:
            has_fields = True
            fields = namespace.pop(FIELDS_META_ATTR)
            field_sedes = tuple(sedes for field_name, sedes in fields)
            try:
                sedes = Container(field_sedes)
            except ValidationError as exception:
                # catch empty or duplicate fields and reraise as a TypeError as this would be an
                # invalid class definition
                raise TypeError(str(exception)) from exception

        else:
            serializable_bases = tuple(
                base for base in bases if isinstance(base, MetaSerializable)
            )
            bases_with_fields = tuple(
                base for base in serializable_bases if base._meta.has_fields
            )

            if len(bases_with_fields) == 0:
                has_fields = False
                fields = None
                sedes = None
            elif len(bases_with_fields) == 1:
                has_fields = True
                fields = bases_with_fields[0]._meta.fields
                sedes = bases_with_fields[0]._meta.container_sedes
            else:
                raise TypeError(
                    "Fields need to be declared explicitly as class has multiple `Serializable` "
                    "parents with fields themselves"
                )

        # create the class without any fields as neither the class itself nor any of its ancestors
        # have defined fields
        if not has_fields:
            meta = Meta(
                has_fields=False,
                fields=None,
                container_sedes=None,
                field_names=None,
                field_attrs=None,
            )
            return super().__new__(mcls, name, bases, assoc(namespace, "_meta", meta))

        # from here on, we can assume that we've got fields and a sedes object
        if sedes is None:
            raise Exception("Invariant: sedes has been initialized earlier")
        if len(fields) == 0:
            raise Exception(
                "Invariant: number of fields has been checked at initializion of sedes"
            )

        field_names, _ = zip(*fields)
        _validate_field_names(field_names)

        # the actual field values are stored in separate *private* attributes.
        # This computes attribute names that don't conflict with other
        # attributes already present on the class.
        reserved_namespace = set(namespace.keys()).union(
            attr
            for base in bases
            for parent_cls in base.__mro__
            for attr in _get_class_namespace(parent_cls)
        )
        field_attrs = _mk_field_attrs(field_names, reserved_namespace)
        field_props = _mk_field_props(field_names, field_attrs)

        # construct the Meta object to store field information for the class
        meta = Meta(
            has_fields=True,
            fields=fields,
            container_sedes=sedes,
            field_names=field_names,
            field_attrs=field_attrs,
        )
        return super().__new__(
            mcls, name, bases, merge(namespace, field_props, {"_meta": meta})
        )

    #
    # Implement BaseSedes methods as pass-throughs to the container sedes
    #
    def serialize(cls: Type[TSerializable], value: TSerializable) -> bytes:
        # return cls._meta.container_sedes.serialize(value)
        if value._serialize_cache is None:
            value._serialize_cache = cls._meta.container_sedes.serialize(value)
        return value._serialize_cache

    def deserialize(cls: Type[TSerializable], data: bytes) -> TSerializable:
        deserialized_fields = cls._meta.container_sedes.deserialize(data)
        deserialized_field_dict = dict(zip(cls._meta.field_names, deserialized_fields))
        return cls(**deserialized_field_dict)

    def get_hash_tree_root(
        cls: Type[TSerializable], value: TSerializable, cache: bool = True
    ) -> bytes:
        if cache:
            root, cache = cls._meta.container_sedes.get_hash_tree_root_and_leaves(
                value, value.cache
            )
            value.cache = cache
            return root
        else:
            return cls._meta.container_sedes.get_hash_tree_root(value)

    @property
    def is_fixed_sized(cls):
        return cls._meta.container_sedes.is_fixed_sized

    def get_fixed_size(cls):
        return cls._meta.container_sedes.get_fixed_size()


# Make any class created with MetaSerializable an instance of BaseSedes
BaseSedes.register(MetaSerializable)


class Serializable(BaseSerializable, metaclass=MetaSerializable):
    """
    The base class for serializable objects.
    """

    def __str__(self) -> str:
        return f"[{', '.join((str(v) for v in self))}]"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {str(self)}>"



import functools
from typing import Any, Iterable

from eth_typing import Hash32
from eth_utils import to_tuple

from ssz.sedes.base import TSedes
from ssz.typing import CacheObj


def get_key(sedes, value: Any) -> str:
    key = get_base_key(sedes, value).hex()
    if len(key) == 0:
        key = ""
    return f"{sedes.get_sedes_id()}{key}"


@functools.lru_cache(maxsize=2 ** 12)
def get_base_key(sedes: TSedes, value: Any) -> bytes:
    return sedes.serialize(value)


@to_tuple
def get_merkle_leaves_without_cache(
    value: Any, element_sedes: TSedes
) -> Iterable[Hash32]:
    for element in value:
        yield element_sedes.get_hash_tree_root(element)


@to_tuple
def get_merkle_leaves_with_cache(
    value: Any, element_sedes: TSedes, cache: CacheObj
) -> Iterable[Hash32]:
    """
    NOTE: cache is mutable
    """
    for element in value:
        key = element_sedes.get_key(element)
        if key not in cache:
            root, cache = element_sedes.get_hash_tree_root_and_leaves(element, cache)
            cache[key] = root
        yield cache[key]



from abc import abstractmethod
import io
import operator
from typing import IO, Any, Generator, Iterable, Sequence, Tuple

from eth_typing import Hash32
from eth_utils import to_tuple
from eth_utils.toolz import accumulate, concatv

from ssz import constants
from ssz.cache.utils import get_key
from ssz.constants import CHUNK_SIZE
from ssz.exceptions import DeserializationError
from ssz.sedes.base import (
    BaseBitfieldCompositeSedes,
    BaseProperCompositeSedes,
    BaseSedes,
    TSedes,
)
from ssz.typing import CacheObj, TDeserialized, TSerializable
from ssz.utils import encode_offset, merkleize, merkleize_with_cache, pack


class BasicSedes(BaseSedes[TSerializable, TDeserialized]):
    def __init__(self, size: int):
        if size <= 0:
            raise ValueError("Length must be greater than 0")

        self.size = size

    #
    # Size
    #
    is_fixed_sized = True

    def get_fixed_size(self):
        return self.size

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: TSerializable) -> Hash32:
        serialized_value = self.serialize(value)
        return merkleize(pack((serialized_value,)))

    def get_hash_tree_root_and_leaves(
        self, value: TSerializable, cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        serialized_value = self.serialize(value)
        return merkleize_with_cache(pack((serialized_value,)), cache=cache)

    def get_key(self, value: Any) -> str:
        return get_key(self, value)


def _compute_fixed_size_section_length(element_sedes: Iterable[TSedes]) -> int:
    return sum(
        sedes.get_fixed_size() if sedes.is_fixed_sized else constants.OFFSET_SIZE
        for sedes in element_sedes
    )


class BitfieldCompositeSedes(BaseBitfieldCompositeSedes[TSerializable, TDeserialized]):
    def get_key(self, value: Any) -> str:
        return get_key(self, value)


class ProperCompositeSedes(BaseProperCompositeSedes[TSerializable, TDeserialized]):
    @to_tuple
    def _get_item_sedes_pairs(
        self, value: Sequence[TSerializable]
    ) -> Generator[Tuple[TSerializable, TSedes], None, None]:
        for index, element in enumerate(value):
            yield element, self.get_element_sedes(index)

    def _validate_serializable(self, value: Any) -> None:
        ...

    def serialize(self, value: TSerializable) -> bytes:
        self._validate_serializable(value)

        if not len(value):
            return b""

        pairs = self._get_item_sedes_pairs(value)  # slow
        element_sedes = tuple(sedes for element, sedes in pairs)

        has_fixed_size_section_length_cache = hasattr(
            value, "_fixed_size_section_length_cache"
        )
        if has_fixed_size_section_length_cache:
            if value._fixed_size_section_length_cache is None:
                fixed_size_section_length = _compute_fixed_size_section_length(
                    element_sedes
                )
                value._fixed_size_section_length_cache = fixed_size_section_length
            else:
                fixed_size_section_length = value._fixed_size_section_length_cache
        else:
            fixed_size_section_length = _compute_fixed_size_section_length(
                element_sedes
            )

        variable_size_section_parts = tuple(
            sedes.serialize(item)  # slow
            for item, sedes in pairs
            if not sedes.is_fixed_sized
        )

        if variable_size_section_parts:
            offsets = tuple(
                accumulate(
                    operator.add,
                    map(len, variable_size_section_parts[:-1]),
                    fixed_size_section_length,
                )
            )
        else:
            offsets = ()

        offsets_iter = iter(offsets)

        fixed_size_section_parts = tuple(
            sedes.serialize(item)  # slow
            if sedes.is_fixed_sized
            else encode_offset(next(offsets_iter))
            for item, sedes in pairs
        )

        try:
            next(offsets_iter)
        except StopIteration:
            pass
        else:
            raise DeserializationError(
                "Did not consume all offsets while decoding value"
            )

        return b"".join(concatv(fixed_size_section_parts, variable_size_section_parts))

    def serialize_element_for_tree(self, index: int, element: TSerializable) -> bytes:
        sedes = self.get_element_sedes(index)
        if self.is_packing:
            return sedes.serialize(element)
        else:
            return sedes.get_hash_tree_root(element)

    @property
    def element_size_in_tree(self) -> int:
        if self.is_packing:
            # note that only homogenous composites with fixed sized elements are packed
            return self.get_element_sedes(0).get_fixed_size()
        else:
            return CHUNK_SIZE

    def deserialize(self, data: bytes) -> TDeserialized:
        stream = io.BytesIO(data)
        value = self._deserialize_stream(stream)
        extra_data = stream.read()
        if extra_data:
            raise DeserializationError(f"Got {len(extra_data)} superfluous bytes")
        return value

    @abstractmethod
    def _deserialize_stream(self, stream: IO[bytes]) -> TDeserialized:
        ...

    def get_key(self, value: Any) -> str:
        return get_key(self, value)


class HomogeneousProperCompositeSedes(
    ProperCompositeSedes[TSerializable, TDeserialized]
):
    def __init__(self, element_sedes: TSedes, max_length: int) -> None:
        self.element_sedes = element_sedes
        if max_length < 1:
            raise ValueError(
                f"(Maximum) length of homogenous composites must be at least 1, got {max_length}"
            )
        self.max_length = max_length

    def get_sedes_id(self) -> str:
        sedes_name = self.__class__.__name__
        return f"{sedes_name}({self.element_sedes.get_sedes_id()},{self.max_length})"

    @property
    def is_packing(self) -> bool:
        return isinstance(self.element_sedes, BasicSedes)

    @property
    def chunk_count(self) -> int:
        if self.is_packing:
            element_size = self.element_sedes.get_fixed_size()
            return (element_size * self.max_length + CHUNK_SIZE - 1) // CHUNK_SIZE
        else:
            return self.max_length



from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, Iterator, Optional, TypeVar, Union

from eth_typing import Hash32
from pyrsistent.typing import PVector

from ssz.hash_tree import HashTree
from ssz.sedes.base import BaseProperCompositeSedes

TStructure = TypeVar("TStructure")
TElement = TypeVar("TElement")


class HashableStructureAPI(ABC, Generic[TElement]):
    @classmethod
    @abstractmethod
    def from_iterable_and_sedes(
        cls,
        iterable: Iterable[TElement],
        sedes: BaseProperCompositeSedes,
        max_length: Optional[int],
    ):
        ...

    #
    # Element and hash tree access
    #
    @property
    @abstractmethod
    def elements(self) -> PVector[TElement]:
        ...

    @property
    @abstractmethod
    def chunks(self) -> PVector[Hash32]:
        ...

    @property
    @abstractmethod
    def hash_tree(self) -> HashTree:
        ...

    @property
    @abstractmethod
    def raw_root(self) -> Hash32:
        ...

    @property
    @abstractmethod
    def hash_tree_root(self) -> Hash32:
        ...

    #
    # Partial PVector interface
    #
    @abstractmethod
    def __len__(self) -> int:
        ...

    @abstractmethod
    def __getitem__(self, index: int) -> TElement:
        ...

    @abstractmethod
    def __iter__(self) -> Iterator[TElement]:
        ...

    @abstractmethod
    def __hash__(self) -> int:
        ...

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        ...

    @abstractmethod
    def transform(self, *transformations):
        ...

    @abstractmethod
    def mset(self: TStructure, *args: Union[int, TElement]) -> TStructure:
        ...

    @abstractmethod
    def set(self: TStructure, index: int, value: TElement) -> TStructure:
        ...

    @abstractmethod
    def evolver(
        self: TStructure
    ) -> "HashableStructureEvolverAPI[TStructure, TElement]":
        ...


class ResizableHashableStructureAPI(HashableStructureAPI[TElement]):
    @abstractmethod
    def append(self: TStructure, value: TElement) -> TStructure:
        ...

    @abstractmethod
    def extend(self: TStructure, values: Iterable[TElement]) -> TStructure:
        ...

    @abstractmethod
    def __add__(self: TStructure, values: Iterable[TElement]) -> TStructure:
        ...

    @abstractmethod
    def __mul__(self: TStructure, times: int) -> TStructure:
        ...

    @abstractmethod
    def evolver(
        self: TStructure
    ) -> "ResizableHashableStructureEvolverAPI[TStructure, TElement]":
        ...


class HashableStructureEvolverAPI(ABC, Generic[TStructure, TElement]):
    @abstractmethod
    def __init__(self, hashable_structure: TStructure) -> None:
        ...

    @abstractmethod
    def __getitem__(self, index: int) -> TElement:
        ...

    @abstractmethod
    def set(self, index: int, element: TElement) -> None:
        ...

    @abstractmethod
    def __setitem__(self, index: int, element: TElement) -> None:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    @abstractmethod
    def is_dirty(self) -> bool:
        ...

    @abstractmethod
    def persistent(self) -> TStructure:
        ...


class ResizableHashableStructureEvolverAPI(
    HashableStructureEvolverAPI[TStructure, TElement]
):
    @abstractmethod
    def append(self, element: TElement) -> None:
        ...

    @abstractmethod
    def extend(self, iterable: Iterable[TElement]) -> None:
        ...



from collections.abc import Mapping
from typing import Sequence

from eth_utils import to_tuple

from ssz.hashable_container import MetaHashableContainer
from ssz.hashable_list import HashableList
from ssz.hashable_vector import HashableVector
from ssz.sedes import (
    BaseSedes,
    Bitlist,
    Bitvector,
    Boolean,
    ByteVector,
    Container,
    List,
    UInt,
    Vector,
)
from ssz.sedes.serializable import MetaSerializable

from .codec import DefaultCodec


def from_formatted_dict(value, sedes, codec=DefaultCodec):
    return parse(value, sedes, codec)


def parse(value, sedes, codec=DefaultCodec):
    if isinstance(sedes, Boolean):
        return parse_boolean(value, sedes, codec)
    elif isinstance(sedes, UInt):
        return parse_integer(value, sedes, codec)
    elif isinstance(sedes, ByteVector):
        return parse_bytes(value, sedes, codec)
    elif isinstance(sedes, List):
        return parse_list(value, sedes, codec)
    elif isinstance(sedes, Vector):
        return parse_vector(value, sedes, codec)
    elif isinstance(sedes, Container):
        return parse_container(value, sedes, codec)
    elif isinstance(sedes, (Bitlist, Bitvector)):
        return parse_bits(value, sedes, codec)
    elif isinstance(sedes, MetaSerializable):
        return parse_serializable(value, sedes, codec)
    elif isinstance(sedes, MetaHashableContainer):
        return parse_hashable(value, sedes, codec)
    elif isinstance(sedes, BaseSedes):
        raise Exception(
            f"Unreachable: All sedes types have been checked, {sedes} was not found"
        )
    else:
        raise TypeError(f"Expected BaseSedes, got {type(sedes)}")


def parse_boolean(value, sedes, codec):
    if not isinstance(value, bool):
        raise ValueError(f"Expected value of type bool, got {type(value)}")
    return codec.decode_bool(value, sedes)


def parse_integer(value, sedes, codec):
    return codec.decode_integer(value, sedes)


def parse_bytes(value, sedes, codec):
    return codec.decode_bytes(value, sedes)


def parse_list(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    elements = tuple(parse(element, sedes.element_sedes, codec) for element in value)
    return HashableList.from_iterable(elements, sedes=sedes)


def parse_vector(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    if not len(value) == sedes.length:
        raise ValueError(f"Expected {sedes.length} elements, got {len(value)}")
    elements = tuple(parse(element, sedes.element_sedes, codec) for element in value)
    return HashableVector.from_iterable(elements, sedes=sedes)


def parse_bits(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")

    data = parse_bytes(value, sedes, codec)
    deserialized = sedes.deserialize(data)
    return tuple(parse_boolean(element, sedes, codec) for element in deserialized)


@to_tuple
def parse_container(value, sedes, codec):
    if not isinstance(value, Sequence):
        raise ValueError(f"Expected Sequence, got {type(value)}")
    elif not len(value) == len(sedes.field_sedes):
        raise ValueError(
            f"Expected {len(sedes.field_sedes)} elements, got {len(value)}"
        )

    for element, element_sedes in zip(value, sedes.field_sedes):
        yield parse(element, element_sedes, codec)


def parse_serializable(value, serializable_cls, codec):
    if not isinstance(value, Mapping):
        raise ValueError(f"Expected Mapping, got {type(value)}")
    parse_args = tuple(
        value[field_name] for field_name in serializable_cls._meta.field_names
    )
    input_args = parse(parse_args, serializable_cls._meta.container_sedes)
    input_kwargs = dict(zip(serializable_cls._meta.field_names, input_args))
    return serializable_cls(**input_kwargs)


def parse_hashable(value, hashable_cls, codec):
    if not isinstance(value, Mapping):
        raise ValueError(f"Expected Mapping, got {type(value)}")
    input_kwargs = {
        field_name: parse(value[field_name], field_type, codec)
        for field_name, field_type in hashable_cls._meta.fields
    }
    return hashable_cls.create(**input_kwargs)



from typing import TYPE_CHECKING, Iterable, Sequence, TypeVar

from eth_typing import Hash32
from pyrsistent import pvector

from ssz.hashable_structure import BaseHashableStructure

if TYPE_CHECKING:
    from ssz.sedes import Vector

TElement = TypeVar("TElement")


class HashableVector(BaseHashableStructure[TElement], Sequence[TElement]):
    @classmethod
    def from_iterable(
        cls, iterable: Iterable[TElement], sedes: "Vector[TElement, TElement]"
    ):
        elements = pvector(iterable)
        if len(elements) != sedes.length:
            raise ValueError(
                "Vector has length {sedes.length}, but {len(elements)} elements are given"
            )
        return super().from_iterable_and_sedes(elements, sedes, max_length=None)

    @property
    def hash_tree_root(self) -> Hash32:
        return self.raw_root



from typing import IO, Any, Iterable, Sequence, Tuple

from eth_typing import Hash32
from eth_utils import to_tuple
from eth_utils.toolz import sliding_window

from ssz.cache.utils import (
    get_merkle_leaves_with_cache,
    get_merkle_leaves_without_cache,
)
from ssz.exceptions import SerializationError
from ssz.hashable_structure import BaseHashableStructure
from ssz.hashable_vector import HashableVector
from ssz.sedes.base import BaseSedes, TSedes
from ssz.sedes.basic import BasicSedes, HomogeneousProperCompositeSedes
from ssz.typing import CacheObj, TDeserializedElement, TSerializableElement
from ssz.utils import merkleize, merkleize_with_cache, pack, read_exact, s_decode_offset

TSedesPairs = Tuple[
    Tuple[BaseSedes[TSerializableElement, TDeserializedElement], TSerializableElement],
    ...,
]


class Vector(
    HomogeneousProperCompositeSedes[
        Sequence[TSerializableElement], Tuple[TDeserializedElement, ...]
    ]
):
    def __init__(self, element_sedes: TSedes, length: int) -> None:
        if length <= 0:
            raise ValueError(f"Vectors must have a size of 1 or greater, got {length}")
        super().__init__(element_sedes, max_length=length)

    @property
    def length(self) -> int:
        return self.max_length

    def get_element_sedes(
        self, index
    ) -> BaseSedes[TSerializableElement, TDeserializedElement]:
        return self.element_sedes

    #
    # Size
    #
    @property
    def is_fixed_sized(self) -> bool:
        return self.element_sedes.is_fixed_sized

    def get_fixed_size(self) -> int:
        if not self.is_fixed_sized:
            raise ValueError("Vector is not fixed size.")

        if self.length == 0:
            return 0
        else:
            return self.length * self.element_sedes.get_fixed_size()

    #
    # Serialization
    #
    def _validate_serializable(self, value: Any) -> None:
        if len(value) != self.length:
            raise SerializationError(
                f"Length mismatch.  Cannot serialize value with length "
                f"{len(value)} as {self.length}-tuple"
            )

    #
    # Deserialization
    #
    def _deserialize_stream(self, stream: IO[bytes]) -> Iterable[TDeserializedElement]:
        elements = self._deserialize_stream_to_tuple(stream)
        return HashableVector.from_iterable(elements, sedes=self)

    @to_tuple
    def _deserialize_stream_to_tuple(
        self, stream: IO[bytes]
    ) -> Iterable[TDeserializedElement]:
        if self.element_sedes.is_fixed_sized:
            element_size = self.element_sedes.get_fixed_size()
            for _ in range(self.length):
                element_data = read_exact(element_size, stream)
                yield self.element_sedes.deserialize(element_data)
        else:
            offsets = tuple(s_decode_offset(stream) for _ in range(self.length))

            for left_offset, right_offset in sliding_window(2, offsets):
                element_length = right_offset - left_offset
                element_data = read_exact(element_length, stream)
                yield self.element_sedes.deserialize(element_data)

            # simply reading to the end of the current stream gives us all of the final element data
            final_element_data = stream.read()
            yield self.element_sedes.deserialize(final_element_data)

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: Sequence[Any]) -> bytes:
        if isinstance(value, BaseHashableStructure) and value.sedes == self:
            return value.hash_tree_root

        if isinstance(self.element_sedes, BasicSedes):
            serialized_elements = tuple(
                self.element_sedes.serialize(element) for element in value
            )
            return merkleize(pack(serialized_elements))
        else:
            element_tree_hashes = tuple(
                self.element_sedes.get_hash_tree_root(element) for element in value
            )
            return merkleize(element_tree_hashes)

    def get_hash_tree_root_and_leaves(
        self, value: Sequence[Any], cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        merkle_leaves = ()
        if isinstance(self.element_sedes, BasicSedes):
            serialized_elements = tuple(
                self.element_sedes.serialize(element) for element in value
            )
            merkle_leaves = pack(serialized_elements)
        else:
            has_get_hash_tree_root_and_leaves = hasattr(
                self.element_sedes, "get_hash_tree_root_and_leaves"
            )
            if has_get_hash_tree_root_and_leaves:
                merkle_leaves = get_merkle_leaves_with_cache(
                    value, self.element_sedes, cache
                )
            else:
                merkle_leaves = get_merkle_leaves_without_cache(
                    value, self.element_sedes
                )

        return merkleize_with_cache(merkle_leaves, cache=cache, limit=self.chunk_count)

    #
    # Equality and hashing
    #
    def __hash__(self) -> int:
        return hash((hash(Vector), hash(self.element_sedes), self.length))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Vector)
            and other.element_sedes == self.element_sedes
            and other.length == self.length
        )



import collections
import functools
from typing import IO, Any, Sequence, Tuple

from eth_typing import Hash32

from ssz.constants import BASE_TYPES, CHUNK_SIZE, EMPTY_CHUNK, OFFSET_SIZE, ZERO_HASHES
from ssz.exceptions import DeserializationError
from ssz.hash import hash_eth2
from ssz.typing import CacheObj


def get_duplicates(values):
    counts = collections.Counter(values)
    return tuple(item for item, num in counts.items() if num > 1)


def read_exact(num_bytes: int, stream: IO[bytes]) -> bytes:
    data = stream.read(num_bytes)
    if len(data) != num_bytes:
        raise DeserializationError(
            f"Tried to read {num_bytes}. Only got {len(data)} bytes"
        )
    return data


def encode_offset(offset: int) -> bytes:
    return offset.to_bytes(OFFSET_SIZE, "little")


def decode_offset(data: bytes) -> int:
    return int.from_bytes(data, "little")


def s_decode_offset(stream: IO[bytes]) -> int:
    data = read_exact(OFFSET_SIZE, stream)
    return decode_offset(data)


def get_items_per_chunk(item_size: int) -> int:
    if item_size < 0:
        raise ValueError("Item size must be positive integer")
    elif item_size == 0:
        return 1
    elif CHUNK_SIZE % item_size != 0:
        raise ValueError("Item size must be a divisor of chunk size")
    elif item_size <= CHUNK_SIZE:
        return CHUNK_SIZE // item_size
    else:
        raise Exception("Invariant: unreachable")


def pad_zeros(value: bytes) -> bytes:
    if len(value) >= CHUNK_SIZE:
        raise ValueError(
            f"The length of given value {len(value)} should be less than CHUNK_SIZE ({CHUNK_SIZE})"
        )
    return value.ljust(CHUNK_SIZE, b"\x00")


@functools.lru_cache(maxsize=2 ** 12)
def to_chunks(packed_data: bytes) -> Tuple[bytes, ...]:
    size = len(packed_data)
    number_of_full_chunks = size // CHUNK_SIZE
    last_chunk_is_full = size % CHUNK_SIZE == 0

    full_chunks = tuple(
        packed_data[chunk_index * CHUNK_SIZE : (chunk_index + 1) * CHUNK_SIZE]
        for chunk_index in range(number_of_full_chunks)
    )
    if last_chunk_is_full:
        return full_chunks
    else:
        last_chunk = pad_zeros(packed_data[number_of_full_chunks * CHUNK_SIZE :])
        return full_chunks + (last_chunk,)


@functools.lru_cache(maxsize=2 ** 12)
def pack(serialized_values: Sequence[bytes]) -> Tuple[Hash32, ...]:
    if len(serialized_values) == 0:
        return (EMPTY_CHUNK,)

    data = b"".join(serialized_values)
    return to_chunks(data)


@functools.lru_cache(maxsize=2 ** 12)
def pack_bytes(byte_string: bytes) -> Tuple[bytes, ...]:
    if len(byte_string) == 0:
        return (EMPTY_CHUNK,)

    return to_chunks(byte_string)


@functools.lru_cache(maxsize=2 ** 12)
def pack_bits(values: Sequence[bool]) -> Tuple[Hash32]:
    as_bytearray = get_serialized_bytearray(values, len(values), extra_byte=False)
    packed = bytes(as_bytearray)
    return to_chunks(packed)


def get_next_power_of_two(value: int) -> int:
    if value <= 0:
        return 1
    else:
        return 2 ** (value - 1).bit_length()


def _get_chunk_and_max_depth(
    chunks: Sequence[Hash32], limit: int, chunk_len: int
) -> Tuple[int, int]:
    chunk_depth = max(chunk_len - 1, 0).bit_length()
    max_depth = max(chunk_depth, (limit - 1).bit_length())
    if max_depth > len(ZERO_HASHES):
        raise ValueError(f"The number of layers is greater than {len(ZERO_HASHES)}")

    return chunk_depth, max_depth


def _get_merkleized_result(
    chunks: Sequence[Hash32],
    chunk_len: int,
    chunk_depth: int,
    max_depth: int,
    cache: CacheObj,
) -> Tuple[Hash32, CacheObj]:
    merkleized_result_per_layers = [None for _ in range(max_depth + 1)]

    def merge(leaf: bytes, leaf_index: int) -> None:
        node = leaf
        layer = 0
        while True:
            if leaf_index & (1 << layer) == 0:
                if leaf_index == chunk_len and layer < chunk_depth:
                    # Keep going if we are complementing the void to the next power of 2
                    key = node + ZERO_HASHES[layer]
                    if key not in cache:
                        cache[key] = hash_eth2(key)
                    node = cache[key]
                else:
                    break
            else:
                key = merkleized_result_per_layers[layer] + node
                if key not in cache:
                    cache[key] = hash_eth2(key)
                node = cache[key]
            layer += 1

        merkleized_result_per_layers[layer] = node

    # Merge in leaf by leaf.
    for leaf_index in range(chunk_len):
        merge(chunks[leaf_index], leaf_index)

    # Complement with 0 if empty, or if not the right power of 2
    if 1 << chunk_depth != chunk_len:
        merge(ZERO_HASHES[0], chunk_len)

    # The next power of two may be smaller than the ultimate virtual size,
    # complement with zero-hashes at each depth.
    for layer in range(chunk_depth, max_depth):
        key = merkleized_result_per_layers[layer] + ZERO_HASHES[layer]
        if key not in cache:
            cache[key] = hash_eth2(
                merkleized_result_per_layers[layer] + ZERO_HASHES[layer]
            )
        merkleized_result_per_layers[layer + 1] = cache[key]

    root = merkleized_result_per_layers[max_depth]

    return root, cache


def merkleize_with_cache(
    chunks: Sequence[Hash32], cache: CacheObj, limit: int = None
) -> Tuple[Hash32, CacheObj]:
    chunk_len = len(chunks)
    if limit is None:
        limit = chunk_len
    chunk_depth, max_depth = _get_chunk_and_max_depth(chunks, limit, chunk_len)

    if limit == 0:
        return ZERO_HASHES[0], cache

    return _get_merkleized_result(
        chunks=chunks,
        chunk_len=chunk_len,
        chunk_depth=chunk_depth,
        max_depth=max_depth,
        cache=cache,
    )


def merkleize(chunks: Sequence[Hash32], limit: int = None) -> Hash32:
    root, _ = merkleize_with_cache(chunks, {}, limit)
    return root


def mix_in_length(root: Hash32, length: int) -> Hash32:
    return hash_eth2(root + length.to_bytes(CHUNK_SIZE, "little"))


def get_serialized_bytearray(
    value: Sequence[bool], bit_count: int, extra_byte: bool
) -> bytearray:
    if extra_byte:
        # Serialize Bitlist
        as_bytearray = bytearray(bit_count // 8 + 1)
    else:
        # Serialize Bitvector
        as_bytearray = bytearray((bit_count + 7) // 8)

    for i in range(bit_count):
        as_bytearray[i // 8] |= value[i] << (i % 8)
    return as_bytearray


def is_immutable_field_value(value: Any) -> bool:
    return type(value) in BASE_TYPES or (
        isinstance(value, tuple)
        and (len(value) == 0 or is_immutable_field_value(value[0]))
    )



from typing import Any

from eth_typing import Hash32

from ssz.sedes import infer_sedes
from ssz.sedes.base import BaseSedes


def get_hash_tree_root(value: Any, sedes: BaseSedes = None) -> Hash32:
    if sedes is None:
        sedes = infer_sedes(value)

    return sedes.get_hash_tree_root(value)



from typing import TYPE_CHECKING, Iterable, Sequence, TypeVar

from eth_typing import Hash32

from ssz.hashable_structure import BaseResizableHashableStructure
from ssz.utils import mix_in_length

TElement = TypeVar("TElement")

if TYPE_CHECKING:
    from ssz.sedes import List


class HashableList(BaseResizableHashableStructure[TElement], Sequence[TElement]):
    @classmethod
    def from_iterable(
        cls, iterable: Iterable[TElement], sedes: "List[TElement, TElement]"
    ):
        return super().from_iterable_and_sedes(
            iterable, sedes, max_length=sedes.max_length
        )

    @property
    def hash_tree_root(self) -> Hash32:
        return mix_in_length(self.raw_root, len(self))



from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, Tuple

from eth_typing import Hash32

from ssz.typing import CacheObj, TDeserialized, TSerializable


class BaseSedes(ABC, Generic[TSerializable, TDeserialized]):
    #
    # Size
    #
    @property
    @abstractmethod
    def is_fixed_sized(self) -> bool:
        ...

    @abstractmethod
    def get_fixed_size(self) -> int:
        ...

    #
    # Serialization
    #
    @abstractmethod
    def serialize(self, value: TSerializable) -> bytes:
        ...

    #
    # Deserialization
    #
    @abstractmethod
    def deserialize(self, data: bytes) -> TDeserialized:
        ...

    #
    # Tree hashing
    #
    @abstractmethod
    def get_hash_tree_root(self, value: TSerializable) -> Hash32:
        ...

    @abstractmethod
    def get_hash_tree_root_and_leaves(
        self, value: TSerializable, cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        ...

    @abstractmethod
    def get_sedes_id(self) -> str:
        ...

    @abstractmethod
    def get_key(self, value: Any) -> str:
        ...

    #
    # Equality and hashing
    #
    @abstractmethod
    def __hash__(self) -> int:
        ...

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        ...


TSedes = BaseSedes[Any, Any]


class BaseProperCompositeSedes(BaseSedes[TSerializable, TDeserialized]):
    @property
    @abstractmethod
    def is_packing(self) -> bool:
        ...

    @abstractmethod
    def get_element_sedes(self, index: int) -> BaseSedes:
        ...

    @property
    @abstractmethod
    def element_size_in_tree(self) -> int:
        ...

    @abstractmethod
    def serialize_element_for_tree(self, index: int, element: TSerializable) -> bytes:
        ...

    @property
    @abstractmethod
    def chunk_count(self) -> Optional[int]:
        ...


class BaseBitfieldCompositeSedes(BaseSedes[TSerializable, TDeserialized]):
    ...



from eth_typing import Hash32
from eth_utils.toolz import iterate, take

from ssz.hash import hash_eth2

CHUNK_SIZE = 32  # named BYTES_PER_CHUNK in the spec
EMPTY_CHUNK = Hash32(b"\x00" * CHUNK_SIZE)

SIGNATURE_FIELD_NAME = "signature"

# number of bytes for a serialized offset
OFFSET_SIZE = 4

FIELDS_META_ATTR = "fields"

ZERO_BYTES32 = Hash32(b"\x00" * 32)
MAX_ZERO_HASHES_LAYER = 100
ZERO_HASHES = tuple(
    take(
        MAX_ZERO_HASHES_LAYER,
        iterate(lambda child: hash_eth2(child + child), ZERO_BYTES32),
    )
)

BASE_TYPES = (int, bytes, bool)



from typing import Any, Sequence, Tuple, Union

from eth_typing import Hash32
from eth_utils import to_tuple

from ssz.exceptions import DeserializationError, SerializationError
from ssz.sedes.basic import BitfieldCompositeSedes
from ssz.typing import CacheObj
from ssz.utils import (
    get_serialized_bytearray,
    merkleize,
    merkleize_with_cache,
    pack_bits,
)

BytesOrByteArray = Union[bytes, bytearray]


class Bitvector(BitfieldCompositeSedes[BytesOrByteArray, bytes]):
    def __init__(self, bit_count: int) -> None:
        if bit_count < 0:
            raise TypeError("Bit count cannot be negative")
        self.bit_count = bit_count

    #
    # Size
    #
    is_fixed_sized = True

    def get_fixed_size(self):
        return (self.bit_count + 7) // 8

    #
    # Serialization
    #
    def serialize(self, value: Sequence[bool]) -> bytes:
        if len(value) != self.bit_count:
            raise SerializationError(
                f"Cannot serialize length {len(value)} bit array as Bitvector[{self.bit_count}]"
            )
        return bytes(get_serialized_bytearray(value, self.bit_count, extra_byte=False))

    #
    # Deserialization
    #
    @to_tuple
    def deserialize(self, data: bytes) -> bytes:
        if len(data) > (self.bit_count + 7) // 8:
            raise DeserializationError(
                f"Cannot deserialize length {len(data)} bytes data as Bitvector[{self.bit_count}]"
            )

        for bit_index in range(self.bit_count):
            yield bool((data[bit_index // 8] >> bit_index % 8) % 2)

    #
    # Tree hashing
    #
    def get_hash_tree_root(self, value: Sequence[bool]) -> bytes:
        return merkleize(pack_bits(value), limit=self.chunk_count)

    def get_hash_tree_root_and_leaves(
        self, value: Sequence[bool], cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        return merkleize_with_cache(
            pack_bits(value), cache=cache, limit=self.chunk_count
        )

    @property
    def chunk_count(self) -> int:
        return (self.bit_count + 255) // 256

    def get_sedes_id(self) -> str:
        return f"{self.__class__.__name__}{self.bit_count}"

    #
    # Equality and hashing
    #
    def __hash__(self) -> int:
        return hash((hash(Bitvector), hash(self.bit_count)))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Bitvector) and other.bit_count == self.bit_count



from collections.abc import MutableMapping
from typing import NewType, TypeVar, Union

TSerializable = TypeVar("TSerializable")
TDeserialized = TypeVar("TDeserialized")

TSerializableElement = TypeVar("TSerializable")
TDeserializedElement = TypeVar("TDeserialized")

CacheObj = NewType("CacheObj", Union[MutableMapping, dict])



from eth_utils import is_bytes

from ssz.sedes import infer_sedes, sedes_by_name
from ssz.sedes.base import BaseSedes


def encode(value, sedes=None):
    """
    Encode object in SSZ format.
    `sedes` needs to be explicitly mentioned for encode/decode
    of integers(as of now).
    `sedes` parameter could be given as a string or as the
    actual sedes object itself.
    """
    if sedes is not None:
        if sedes in sedes_by_name:
            # Get the actual sedes object from string representation
            sedes_obj = sedes_by_name[sedes]
        else:
            sedes_obj = sedes

        if not isinstance(sedes_obj, BaseSedes):
            raise TypeError("Invalid sedes object")

    else:
        sedes_obj = infer_sedes(value)

    return sedes_obj.serialize(value)


def decode(ssz, sedes):
    """
    Decode a SSZ encoded object.
    """
    if not is_bytes(ssz):
        raise TypeError(f"Can only decode SSZ bytes, got type {type(ssz).__name__}")

    value = sedes.deserialize(ssz)
    return value



from collections.abc import Iterable

from ssz.abc import HashableStructureAPI

from .base import BaseSedes  # noqa: F401
from .basic import BasicSedes, ProperCompositeSedes  # noqa: F401
from .bitlist import Bitlist  # noqa: F401
from .bitvector import Bitvector  # noqa: F401
from .boolean import Boolean, boolean  # noqa: F401
from .byte import Byte, byte  # noqa: F401
from .byte_vector import (  # noqa: F401
    ByteVector,
    bytes1,
    bytes4,
    bytes32,
    bytes48,
    bytes96,
)
from .container import Container  # noqa: F401
from .list import List  # noqa: F401
from .serializable import Serializable  # noqa: F401
from .signed_serializable import SignedSerializable  # noqa: F401
from .uint import UInt, uint8, uint16, uint32, uint64, uint128, uint256  # noqa: F401
from .vector import Vector  # noqa: F401

sedes_by_name = {
    "bool": boolean,
    "byte": byte,
    "bytes4": bytes4,
    "bytes32": bytes32,
    "bytes48": bytes48,
    "bytes96": bytes96,
    "uint8": uint8,
    "uint16": uint16,
    "uint32": uint32,
    "uint64": uint64,
    "uint128": uint128,
    "uint256": uint256,
}


def infer_sedes(value):
    """
    Try to find a sedes objects suitable for a given Python object.
    """
    if isinstance(value.__class__, BaseSedes):
        return value.__class__
    elif isinstance(value, HashableStructureAPI):
        return value.sedes
    elif isinstance(value, bool):
        return boolean
    elif isinstance(value, int):
        raise TypeError(
            "uint sedes object or uint string needs to be specified for ints"
        )
    elif isinstance(value, Iterable):
        raise TypeError("Cannot infer list sedes for iterables that are not sequences")
    else:
        raise TypeError(f"Did not find sedes handling type {type(value).__name__}")



from abc import ABCMeta
import math
import sys
from typing import (
    Any,
    Dict,
    Generator,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from eth_typing import Hash32
from eth_utils import to_dict, to_tuple
from eth_utils.toolz import merge

from ssz.constants import FIELDS_META_ATTR, SIGNATURE_FIELD_NAME, ZERO_HASHES
from ssz.hashable_list import HashableList
from ssz.hashable_structure import BaseHashableStructure, HashableStructureEvolver
from ssz.hashable_vector import HashableVector
from ssz.sedes import ByteVector, List, Vector
from ssz.sedes.base import BaseSedes
from ssz.sedes.container import Container

TStructure = TypeVar("TStructure", bound="HashableContainer")
TElement = TypeVar("TElement")


class FieldDescriptor:
    """Descriptor translating from __getattr__ to __getitem__ calls for a given attribute."""

    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(
        self, instance: "HashableContainer", owner: Type["HashableContainer"] = None
    ) -> Any:
        return instance[self.name]


class SettableFieldDescriptor(FieldDescriptor):
    """Settable variant of FieldDescriptor for evolver."""

    def __set__(self, instance: "HashableContainerEvolver", value: Any) -> None:
        instance[self.name] = value


Field = Tuple[str, BaseSedes]


class Meta(NamedTuple):
    fields: Tuple[Field, ...]
    field_names: Tuple[str, ...]
    field_names_to_element_indices: Dict[str, int]
    container_sedes: Container
    evolver_class: Type["HashableContainerEvolver"]

    @classmethod
    def from_fields(
        cls, fields: Tuple[Tuple[str, BaseSedes], ...], container: Container, name: str
    ) -> "Meta":
        if not fields:
            raise TypeError("Containers must define at least one field")

        field_names, _ = zip(*fields)
        field_names_to_element_indices = {
            field_name: index for index, field_name in enumerate(field_names)
        }

        # create subclass of HashableEvolver that has a settable descriptor for each field
        settable_field_descriptors = {
            field_name: SettableFieldDescriptor(field_name)
            for field_name in field_names
        }
        evolver_class = type(
            name + "Evolver", (HashableContainerEvolver,), settable_field_descriptors
        )

        return cls(
            fields=fields,
            field_names=field_names,
            field_names_to_element_indices=field_names_to_element_indices,
            container_sedes=container,
            evolver_class=evolver_class,
        )


def get_meta_from_bases(bases: Tuple[Type, ...]) -> Optional[Meta]:
    """Return the meta object defined by one of the given base classes.

    Returns None if no base defines a meta object. Raises a TypeError if more than one do.
    """
    container_bases = tuple(
        base for base in bases if isinstance(base, MetaHashableContainer)
    )
    bases_with_meta = tuple(base for base in container_bases if base._meta is not None)

    if len(bases_with_meta) == 0:
        return None
    elif len(bases_with_meta) == 1:
        return bases_with_meta[0]._meta
    else:
        raise TypeError(
            "Fields need to be declared explicitly as class has multiple "
            "HashableContainer parents with fields themselves"
        )


@to_tuple
def get_field_sedes_from_fields(
    fields: Sequence[Field]
) -> Generator[BaseSedes, None, None]:
    for _, field in fields:
        if isinstance(field, BaseSedes):
            yield field
        else:
            raise TypeError(
                f"Field must either be a sedes object or a hashable container, got {field}"
            )


class MetaHashableContainer(ABCMeta):
    """Metaclass which creates HashableContainers."""

    def __new__(mcls, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any]):
        container_sedes: Optional[Container]

        # get fields defined in the class or by one of its bases
        if FIELDS_META_ATTR in namespace:
            fields = namespace.pop(FIELDS_META_ATTR)
            field_sedes = get_field_sedes_from_fields(fields)
            if not fields:
                raise TypeError(
                    "HashableContainer must either define a non-zero number of fields or not "
                    "define any, but not an empty set."
                )
            container_sedes = Container(field_sedes)
        else:
            meta = get_meta_from_bases(bases)
            if meta is not None:
                fields = meta.fields
                field_sedes = get_field_sedes_from_fields(fields)
                container_sedes = Container(field_sedes)
            else:
                fields = None
                container_sedes = None

        if fields is not None:
            assert container_sedes is not None
            meta = Meta.from_fields(fields, container_sedes, name=name)
            field_descriptors = {
                field_name: FieldDescriptor(field_name)
                for field_name in meta.field_names
            }
        else:
            meta = None
            field_descriptors = {}

        namespace_with_meta_and_fields = merge(
            namespace, {"_meta": meta}, field_descriptors
        )
        return super().__new__(mcls, name, bases, namespace_with_meta_and_fields)

    #
    # Sedes interface
    #
    @property
    def is_fixed_sized(cls):
        return cls._meta.container_sedes.is_fixed_sized

    def get_fixed_size(cls):
        return cls._meta.container_sedes.get_fixed_size()

    def serialize(cls, value):
        return cls._meta.container_sedes.serialize(value)

    def deserialize(cls, data):
        field_values = cls._meta.container_sedes.deserialize(data)
        kwargs = {
            field_name: field_value
            for (field_name, _), field_value in zip(cls._meta.fields, field_values)
        }
        return cls.create(**kwargs)

    def get_hash_tree_root(cls, value):
        return cls._meta.container_sedes.get_hash_tree_root(value)

    def get_hash_tree_root_and_leaves(cls, value, cache):
        return cls._meta.container_sedes.get_hash_tree_root_and_leaves(value, cache)

    def get_sedes_id(cls):
        return cls._meta.container_sedes.get_sedes_id()

    def get_key(cls, value):
        return cls._meta.container_sedes.get_key(value)


class MetaSignedHashableContainer(MetaHashableContainer):
    def __new__(mcls, name, bases, namespace):
        cls = super().__new__(mcls, name, bases, namespace)

        # check that the signature conventions are abided by
        if cls._meta is not None:
            if len(cls._meta.fields) < 2:
                raise TypeError(f"Signed containers need to have at least two fields")
            if cls._meta.fields[-1][0] != SIGNATURE_FIELD_NAME:
                raise TypeError(
                    f"Last field of signed container must be {SIGNATURE_FIELD_NAME}, but is "
                    f"{cls._meta.fields[-1][0]}"
                )

        return cls


BaseSedes.register(MetaHashableContainer)
BaseSedes.register(MetaSignedHashableContainer)


# workaround for https://github.com/python/typing/issues/449 (fixed in python 3.7)
python_version_info = sys.version_info
if python_version_info[0] <= 3 and python_version_info[1] <= 6:
    from typing import GenericMeta

    class GenericMetaHashableContainer(GenericMeta, MetaHashableContainer):
        pass

    class GenericMetaSignedHashableContainer(
        GenericMetaHashableContainer, MetaSignedHashableContainer
    ):
        pass


else:
    GenericMetaHashableContainer = MetaHashableContainer  # type: ignore
    GenericMetaSignedHashableContainer = MetaSignedHashableContainer  # type: ignore


def hashablify_value(value: Any, sedes: BaseSedes) -> Any:
    if isinstance(sedes, List):
        return HashableList.from_iterable(value, sedes)
    elif isinstance(sedes, Vector):
        if isinstance(sedes, ByteVector):
            return value
        else:
            return HashableVector.from_iterable(value, sedes)
    else:
        return value


@to_dict
def hashablify_field_kwargs(
    field_kwargs: Dict[str, Any], fields: Sequence[Field]
) -> Generator[Tuple[str, Any], None, None]:
    for field_name, field_sedes in fields:
        field_value = field_kwargs[field_name]
        hashablified_field_value = hashablify_value(field_value, field_sedes)
        yield field_name, hashablified_field_value


class HashableContainer(
    BaseHashableStructure[TElement], metaclass=GenericMetaHashableContainer
):
    """Base class for hashable containers."""

    _meta: Meta  # set by MetaHashableContainer

    def __init__(self, *args, **kwargs):
        if self._meta is None:
            raise TypeError("HashableContainer does not define any fields")
        else:
            super().__init__(*args, **kwargs)

    @classmethod
    def create(cls, **field_kwargs: Dict[str, Any]):
        if cls._meta is None:
            raise TypeError("HashableContainer does not define any fields")

        given_keys = set(field_kwargs.keys())
        expected_keys = set(cls._meta.field_names_to_element_indices.keys())
        missing_keys = expected_keys - given_keys
        unexpected_keys = given_keys - expected_keys

        if missing_keys:
            raise ValueError(
                f"The following keyword arguments are missing: {', '.join(sorted(missing_keys))}"
            )
        if unexpected_keys:
            raise ValueError(
                f"The following keyword arguments are unexpected: "
                f"{', '.join(sorted(unexpected_keys))}"
            )

        hashablified_field_kwargs = hashablify_field_kwargs(
            field_kwargs, cls._meta.fields
        )
        field_values = (
            hashablified_field_kwargs[field_name]
            for field_name in cls._meta.field_names
        )

        return cls.from_iterable_and_sedes(
            field_values, sedes=cls._meta.container_sedes, max_length=None
        )

    @property
    def hash_tree_root(self) -> Hash32:
        return self.raw_root

    def normalize_item_index(self, index: Union[str, int]) -> int:
        if isinstance(index, str):
            return self._meta.field_names_to_element_indices[index]
        elif isinstance(index, int):
            return index
        else:
            raise TypeError("Index must be either int or str")

    def __getitem__(self, index: Union[str, int]) -> TElement:
        element_index = self.normalize_item_index(index)
        return super().__getitem__(element_index)

    def evolver(self: TStructure) -> "HashableContainerEvolver[TStructure, TElement]":
        return self._meta.evolver_class(self)


class HashableContainerEvolver(HashableStructureEvolver[TStructure, TElement]):
    """Base class for evolvers for hashable containers.

    Subclasses (created dynamically by MetaHashableContainer when creating the corresponding
    HashableContainer) should add settable field descriptors for all fields.
    """

    def __setitem__(self, index: Union[str, int], value: TElement) -> None:
        element_index = self._original_structure.normalize_item_index(index)

        # provides some safety by preventing overwriting a hashable list/vector with another kind
        # of sequence
        container_sedes = self._original_structure._meta.container_sedes
        field_sedes = container_sedes.field_sedes[element_index]
        hashablified_value = hashablify_value(value, field_sedes)

        super().__setitem__(element_index, hashablified_value)

    def __getitem__(self, index: Union[str, int]) -> TElement:
        element_index = self._original_structure.normalize_item_index(index)
        return super().__getitem__(element_index)


class SignedHashableContainer(
    HashableContainer[TElement], metaclass=GenericMetaSignedHashableContainer
):
    _meta: Meta

    @property
    def signing_root(self) -> Hash32:
        signature_chunk_index = len(self) - 1
        hash_tree_with_blank_signature = self._hash_tree.set(
            signature_chunk_index, ZERO_HASHES[0]
        )
        if math.log2(len(self) - 1).is_integer():
            root_layer_index = -2
        else:
            root_layer_index = -1
        return hash_tree_with_blank_signature.raw_hash_tree[root_layer_index][0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

extras_require = {
    "test": [
        "pytest==4.4.0",
        "pytest-xdist==1.28.0",
        "pytest-watch>=4.1.0,<5",
        "tox>=2.9.1,<3",
        "hypothesis==3.69.5",
        "ruamel.yaml==0.15.87",
        "mypy-extensions>=0.4.1,<1.0.0",
    ],
    "lint": ["flake8==3.7.8", "isort==4.3.21", "black==19.3b"],
    "doc": ["Sphinx>=1.6.5,<2", "sphinx_rtd_theme>=0.1.9"],
    "dev": ["bumpversion>=0.5.3,<1", "wheel", "twine", "ipython", "pre-commit==1.18.3"],
}

extras_require["dev"] = (
    extras_require["dev"]
    + extras_require["test"]
    + extras_require["lint"]
    + extras_require["doc"]
)

setup(
    name="ssz",
    # *IMPORTANT*: Don't manually change the version here. Use `make bump`, as described in readme
    version="0.2.1",
    description="""ssz: Python implementation of the Simple Serialization encoding and decoding""",
    long_description_markdown_filename="README.md",
    author="Jason Carver",
    author_email="ethcalibur+pip@gmail.com",
    url="https://github.com/ethereum/py-ssz",
    include_package_data=True,
    install_requires=[
        "eth-utils>=1,<2",
        "lru-dict>=1.1.6",
        # When updating to a newer version of pyrsistent, please check that the interface
        # `transform` expects has not changed (see https://github.com/tobgu/pyrsistent/issues/180)
        "pyrsistent==0.15.6",
    ],
    setup_requires=["setuptools-markdown"],
    python_requires=">=3.6, <4",
    extras_require=extras_require,
    py_modules=["ssz"],
    license="MIT",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



class SSZException(Exception):
    """
    Base class for exceptions raised by this package.
    """

    pass


class SerializationError(SSZException):
    """
    Exception raised if serialization fails.
    """

    pass


class DeserializationError(SSZException):
    """
    Exception raised if deserialization fails.
    """

    pass



from typing import Tuple, Union

from eth_typing import Hash32

from ssz.exceptions import DeserializationError, SerializationError
from ssz.sedes.byte import byte
from ssz.sedes.vector import Vector
from ssz.typing import CacheObj
from ssz.utils import merkleize, merkleize_with_cache, pack_bytes

BytesOrByteArray = Union[bytes, bytearray]


class ByteVector(Vector[BytesOrByteArray, bytes]):
    """Equivalent to `Vector(byte, size)` but more efficient."""

    def __init__(self, size: int) -> None:
        super().__init__(element_sedes=byte, length=size)

    def serialize(self, value: BytesOrByteArray) -> bytes:
        if len(value) != self.length:
            raise SerializationError(
                f"Cannot serialize length {len(value)} byte-string as bytes{self.length}"
            )

        return value

    def serialize_element_for_tree(self, index: int, byte_value: int) -> bytes:
        if not 0 <= byte_value < 256:
            raise SerializationError(
                f"Cannot serialize byte as its value {byte_value} is invalid"
            )

        return bytes((byte_value,))

    def deserialize(self, data: bytes) -> bytes:
        if len(data) != self.length:
            raise DeserializationError(
                f"Cannot deserialize length {len(data)} data as bytes{self.length}"
            )
        return data

    def get_hash_tree_root(self, value: bytes) -> bytes:
        serialized_value = self.serialize(value)
        return merkleize(pack_bytes(serialized_value))

    def get_hash_tree_root_and_leaves(
        self, value: bytes, cache: CacheObj
    ) -> Tuple[Hash32, CacheObj]:
        serialized_value = self.serialize(value)
        return merkleize_with_cache(pack_bytes(serialized_value), cache=cache)

    def get_sedes_id(self) -> str:
        return f"{self.__class__.__name__}{self.length}"


bytes1 = ByteVector(1)
bytes4 = ByteVector(4)
bytes8 = ByteVector(8)
bytes32 = ByteVector(32)
bytes48 = ByteVector(48)
bytes96 = ByteVector(96)



from functools import partial
import itertools
from numbers import Integral
from typing import Any, Generator, Iterable, Optional, Union

# `transform` comes from a non-public API which is considered stable, but future changes can not be
# ruled out completely. Therefore, the implementation should be reviewed whenever pyrsistent is
# updated. See https://github.com/tobgu/pyrsistent/issues/180 for more information.
from eth_typing import Hash32
from eth_utils.toolz import drop, iterate, partition, pipe, take
from pyrsistent import pmap, pvector
from pyrsistent._transformations import transform
from pyrsistent.typing import PMap, PVector

from ssz.constants import ZERO_HASHES
from ssz.hash import hash_eth2
from ssz.utils import get_next_power_of_two

RawHashTreeLayer = PVector[Hash32]
RawHashTree = PVector[RawHashTreeLayer]


def validate_chunk_count(chunk_count: Optional[int]) -> None:
    if chunk_count is not None:
        if chunk_count <= 0:
            raise ValueError(f"Chunk count is not positive: {chunk_count}")


def validate_raw_hash_tree(
    raw_hash_tree: RawHashTree, chunk_count: Optional[int] = None
) -> None:
    if len(raw_hash_tree) == 0:
        raise ValueError("Hash tree is empty")

    if len(raw_hash_tree[0]) == 0:
        raise ValueError("Hash tree contains zero chunks")

    if chunk_count is not None and len(raw_hash_tree[0]) > chunk_count:
        raise ValueError(
            f"Hash tree contains {len(raw_hash_tree[0])} chunks which exceeds chunk count "
            f"{chunk_count}"
        )

    if len(raw_hash_tree[-1]) != 1:
        raise ValueError(
            f"Hash tree root layer contains {len(raw_hash_tree[-1])} items instead of 1"
        )


class HashTree(PVector[Hash32]):
    def __init__(
        self, raw_hash_tree: RawHashTree, chunk_count: Optional[int] = None
    ) -> None:
        validate_chunk_count(chunk_count)
        validate_raw_hash_tree(raw_hash_tree, chunk_count)

        self.chunk_count = chunk_count
        self.raw_hash_tree = raw_hash_tree

    @classmethod
    def compute(
        cls, chunks: Iterable[Hash32], chunk_count: Optional[int] = None
    ) -> "HashTree":
        raw_hash_tree = compute_hash_tree(chunks, chunk_count)
        return cls(raw_hash_tree, chunk_count)

    @property
    def chunks(self) -> RawHashTreeLayer:
        return self.raw_hash_tree[0]

    @property
    def root(self) -> Hash32:
        return self.raw_hash_tree[-1][0]

    def transform(self, *transformations):
        return transform(self, transformations)

    def evolver(self):
        return HashTreeEvolver(self)

    #
    # Comparison
    #
    def __hash__(self) -> int:
        return hash((self.root, self.chunk_count))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, HashTree)
            and self.root == other.root
            and self.chunk_count == other.chunk_count
        )

    #
    # Chunk access
    #
    def __len__(self) -> int:
        return len(self.chunks)

    def __getitem__(self, index: Union[int, slice]) -> Hash32:
        return self.chunks[index]

    def index(self, value: Hash32, *args, **kwargs) -> Hash32:
        return self.chunks.index(value, *args, **kwargs)

    def count(self, value: Hash32) -> int:
        return self.chunks.count(value)

    #
    # Tree modifications via evolver
    #
    def append(self, value: Hash32) -> "HashTree":
        evolver = self.evolver()
        evolver.append(value)
        return evolver.persistent()

    def extend(self, value: Iterable[Hash32]) -> "HashTree":
        evolver = self.evolver()
        evolver.extend(value)
        return evolver.persistent()

    def __add__(self, other: Iterable[Hash32]) -> "HashTree":
        return self.extend(other)

    def __mul__(self, times: int) -> "HashTree":
        if times <= 0:
            raise ValueError(f"Multiplier must be greater or equal to 1, got {times}")

        evolver = self.evolver()

        for _ in range(times - 1):
            evolver.extend(self)

        return evolver.persistent()

    def mset(self, *args: Union[int, Hash32]) -> "HashTree":
        if len(args) % 2 != 0:
            raise TypeError(
                f"mset must be called with an even number of arguments, got {len(args)}"
            )

        evolver = self.evolver()
        for index, value in partition(2, args):
            evolver[index] = value
        return evolver.persistent()

    def set(self, index: int, value: Hash32) -> "HashTree":
        return self.mset(index, value)

    #
    # Removal of chunks
    #
    def delete(self, index: int, stop: Optional[int] = None) -> "HashTree":
        if stop is None:
            stop = index + 1
        chunks = self.chunks.delete(index, stop)
        return self.__class__.compute(chunks, self.chunk_count)

    def remove(self, value: Hash32) -> "HashTree":
        chunks = self.chunks.remove(value)
        return self.__class__.compute(chunks, self.chunk_count)


class HashTreeEvolver:
    def __init__(self, hash_tree: "HashTree") -> None:
        self.original_hash_tree = hash_tree
        self.updated_chunks: PMap[int, Hash32] = pmap()
        self.appended_chunks: PVector[Hash32] = pvector()

    #
    # Getters
    #
    def __getitem__(self, index: int) -> Hash32:
        if index < 0:
            index += len(self)

        if index in self.updated_chunks:
            return self.updated_chunks[index]
        else:
            if 0 <= index < len(self.original_hash_tree):
                return self.original_hash_tree[index]
            elif index < len(self):
                index_in_appendix = index - len(self.original_hash_tree)
                return self.appended_chunks[index_in_appendix]
            else:
                raise IndexError(f"Index out of bounds: {index}")

    def __len__(self) -> int:
        return len(self.original_hash_tree) + len(self.appended_chunks)

    def is_dirty(self) -> bool:
        return self.updated_chunks or self.appended_chunks

    #
    # Setters
    #
    def set(self, index: Integral, value: Hash32) -> None:
        self[index] = value

    def __setitem__(self, index: Integral, value: Hash32) -> None:
        if index < 0:
            index += len(self)

        if 0 <= index < len(self.original_hash_tree):
            self.updated_chunks = self.updated_chunks.set(index, value)
        elif index < len(self):
            index_in_appendix = index - len(self.original_hash_tree)
            self.appended_chunks = self.appended_chunks.set(index_in_appendix, value)
        else:
            raise IndexError(f"Index out of bounds: {index}")

    #
    # Length changing modifiers
    #
    def append(self, value: Hash32) -> None:
        self.appended_chunks = self.appended_chunks.append(value)
        self._check_chunk_count()

    def extend(self, values: Iterable[Hash32]) -> None:
        self.appended_chunks = self.appended_chunks.extend(values)
        self._check_chunk_count()

    def _check_chunk_count(self) -> None:
        chunk_count = self.original_hash_tree.chunk_count
        if chunk_count is not None and len(self) > chunk_count:
            raise ValueError("Hash tree exceeds size chunk count {chunk_count}")

    #
    # Not implemented
    #
    def delete(self, index, stop=None):
        raise NotImplementedError()

    def __delitem__(self, index):
        raise NotImplementedError()

    def remove(self, value):
        raise NotImplementedError()

    #
    # Persist
    #
    def persistent(self) -> "HashTree":
        if not self.is_dirty():
            return self.original_hash_tree
        else:
            setters = (
                partial(set_chunk_in_tree, index=index, chunk=chunk)
                for index, chunk in self.updated_chunks.items()
            )
            appenders = (
                partial(append_chunk_to_tree, chunk=chunk)
                for chunk in self.appended_chunks
            )
            raw_hash_tree = pipe(
                self.original_hash_tree.raw_hash_tree, *setters, *appenders
            )
            return self.original_hash_tree.__class__(
                raw_hash_tree, self.original_hash_tree.chunk_count
            )


def hash_layer(child_layer: RawHashTreeLayer, layer_index: int) -> RawHashTreeLayer:
    if len(child_layer) % 2 == 0:
        padded_child_layer = child_layer
    else:
        padded_child_layer = child_layer.append(ZERO_HASHES[layer_index])

    child_pairs = partition(2, padded_child_layer)
    parent_layer = pvector(
        hash_eth2(left_child + right_child) for left_child, right_child in child_pairs
    )
    return parent_layer


def generate_hash_tree_layers(
    chunks: RawHashTreeLayer
) -> Generator[RawHashTreeLayer, None, None]:
    yield chunks
    previous_layer = chunks

    for previous_layer_index in itertools.count():
        # stop if the root has been reached
        if len(previous_layer) <= 1:
            break
        if previous_layer_index > 1000:
            raise Exception("Invariant: Root is reached quickly")

        next_layer = hash_layer(previous_layer, previous_layer_index)

        yield next_layer
        previous_layer = next_layer


def get_num_layers(num_chunks: int, chunk_count: Optional[int]) -> int:
    if chunk_count is not None:
        virtual_size = chunk_count
    else:
        virtual_size = num_chunks

    return get_next_power_of_two(virtual_size).bit_length()


def generate_chunk_tree_padding(
    unpadded_chunk_tree: PVector[Hash32], chunk_count: Optional[int]
) -> Generator[Hash32, None, None]:
    if chunk_count is None:
        return

    num_chunks = len(unpadded_chunk_tree[0])
    if num_chunks > chunk_count:
        raise ValueError(
            f"Number of chunks in tree ({num_chunks}) exceeds chunk count {chunk_count}"
        )

    num_existing_layers = len(unpadded_chunk_tree)
    num_target_layers = get_next_power_of_two(chunk_count).bit_length()

    previous_root = unpadded_chunk_tree[-1][0]
    for previous_layer_index in range(num_existing_layers - 1, num_target_layers - 1):
        next_root = hash_eth2(previous_root + ZERO_HASHES[previous_layer_index])
        yield pvector([next_root])
        previous_root = next_root


def pad_hash_tree(
    unpadded_chunk_tree: RawHashTree, chunk_count: Optional[int] = None
) -> RawHashTree:
    padding = pvector(generate_chunk_tree_padding(unpadded_chunk_tree, chunk_count))
    return unpadded_chunk_tree + padding


def compute_hash_tree(
    chunks: Iterable[Hash32], chunk_count: Optional[int] = None
) -> RawHashTree:
    validate_chunk_count(chunk_count)

    chunks = pvector(chunks)

    if not chunks:
        raise ValueError("Number of chunks is 0")
    if chunk_count is not None and len(chunks) > chunk_count:
        raise ValueError(
            f"Number of chunks ({len(chunks)}) exceeds chunk_count ({chunk_count})"
        )

    unpadded_chunk_tree = pvector(generate_hash_tree_layers(chunks))
    return pad_hash_tree(unpadded_chunk_tree, chunk_count)


def recompute_hash_in_tree(
    hash_tree: RawHashTree, layer_index: int, hash_index: int
) -> RawHashTree:
    if layer_index == 0:
        raise ValueError(
            "Cannot recompute hash in leaf layer as it consists of chunks not hashes"
        )

    child_layer_index = layer_index - 1
    left_child_hash_index = hash_index * 2
    right_child_hash_index = left_child_hash_index + 1

    child_layer = hash_tree[child_layer_index]
    left_child_hash = child_layer[left_child_hash_index]
    try:
        right_child_hash = child_layer[right_child_hash_index]
    except IndexError:
        right_child_hash = ZERO_HASHES[child_layer_index]

    # create the layer if it doesn't exist yet (otherwise, pyrsistent would create a PMap)
    if layer_index == len(hash_tree):
        hash_tree = hash_tree.append(pvector())

    parent_hash = hash_eth2(left_child_hash + right_child_hash)
    return hash_tree.transform((layer_index, hash_index), parent_hash)


def set_chunk_in_tree(hash_tree: RawHashTree, index: int, chunk: Hash32) -> RawHashTree:
    hash_tree_with_updated_chunk = hash_tree.transform((0, index), chunk)

    parent_layer_indices = drop(1, range(len(hash_tree)))
    parent_hash_indices = drop(
        1, take(len(hash_tree), iterate(lambda index: index // 2, index))
    )

    update_functions = (
        partial(recompute_hash_in_tree, layer_index=layer_index, hash_index=hash_index)
        for layer_index, hash_index in zip(parent_layer_indices, parent_hash_indices)
    )

    hash_tree_with_updated_branch = pipe(
        hash_tree_with_updated_chunk, *update_functions
    )

    if len(hash_tree_with_updated_branch[-1]) == 1:
        return hash_tree_with_updated_branch
    elif len(hash_tree_with_updated_branch[-1]) == 2:
        return recompute_hash_in_tree(hash_tree_with_updated_branch, len(hash_tree), 0)
    else:
        raise Exception("Unreachable")


def append_chunk_to_tree(hash_tree: RawHashTree, chunk: Hash32) -> RawHashTree:
    return set_chunk_in_tree(hash_tree, len(hash_tree[0]), chunk)
